"""

Testkit command line tool caller.

"""
import logging
import subprocess
import json
import shutil
import os
from typing import List, Dict, Any


logger = logging.getLogger("metta_testkit:testkit")

TESTKITCLIENT_CLI_CONFIG_FILE_DEFAULT = "./testkit.yml"
""" Testkit config configuration file key """

TESTKITCLIENT_WORKING_DIR_DEFAULT = "."
""" Testkit Client default working dir """
TESTKITCLIENT_BIN_PATH = "testkit"
""" Testkit bin exec for the subprocess """


class TestkitClient:
    """Shell client for interacting with the testkit bin."""

    def __init__(
        self,
        config_file: str = TESTKITCLIENT_CLI_CONFIG_FILE_DEFAULT,
        working_dir: str = TESTKITCLIENT_WORKING_DIR_DEFAULT,
        debug: bool = False,
        binary: str = TESTKITCLIENT_BIN_PATH,
    ):
        """Initialize Testkit command executer.

        Parameters:
        -----------
        config_file (str) : path to testkit config file, typically
            testkit.yml

        debug (bool) : passed to the testkit client to tell it to enable
            verbose debugging output.

        """
        self.config_file = config_file
        """ Path to config file """
        self.working_dir = working_dir
        """ Python subprocess working dir to execute testkit in.
            This may be relevant in cases where ssh keys have a relative path """

        self.debug = debug
        """ should testkit be run with verbose output enabled ? """

        if shutil.which(binary) is None:
            raise ValueError(
                "Testkit binary not found. Testkit commands cannot be called. "
                f"Expected binary at path {binary}"
            )

        self.bin = binary
        """ path to testkit executable """

    # the deep argument is a standard for the info hook
    # pylint: disable=unused-argument
    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Get info about a provisioner plugin."""
        return {
            "config_file": self.config_file,
            "working_dir": self.working_dir,
            "bin": self.bin,
            "version": self.version(),
        }

    def version(self):
        """Return testkit client version."""
        return self._run(["version"], return_output=True, use_config_file=False)

    def create(self, opts: List[str]):
        """Run the testkit create command."""
        return self._run(["create"] + opts)

    def system_ls(self):
        """List all of the systems testkit can see using our config."""
        return json.loads(self._run(["system", "ls", "--json"], return_output=True))

    def system_rm(self, system_name: str):
        """Remove a system from testkit."""
        return self._run(["system", "rm", system_name])

    def machine_ls(self, system_name: str):
        """Remove a system from testkit."""
        return json.loads(
            self._run(
                ["machine", "ls", "--filter", f"name={system_name}", "--json"],
                return_output=True,
            )
        )

    def machine_ssh(self, machine: str, cmd: str = ""):
        """Run an ssh command onto a host."""
        return self._run(["machine", "ssh", machine, cmd])

    # this syntax makes it easier to read
    # pylint: disable=inconsistent-return-statements
    def _run(self, args: List[str], return_output=False, use_config_file: bool = True):
        """Run a testkit command.

        Parameters:
        -----------
        args (List[str]) : arguments to pass to the testkit bin using
            subprocess

        """
        cmd = [self.bin]

        if use_config_file and os.path.isfile(self.config_file):
            cmd += [f"--file={self.config_file}"]

        if self.debug:
            cmd += ["--debug"]

        cmd += args

        if not return_output:
            logger.debug("running testkit command: %s", " ".join(cmd))
            res = subprocess.run(cmd, cwd=self.working_dir, text=True, check=True)
            res.check_returncode()
        else:
            logger.debug("running testkit command with output capture: %s", " ".join(cmd))
            res = subprocess.run(
                cmd,
                cwd=self.working_dir,
                shell=False,
                stdout=subprocess.PIPE,
                check=True,
            )
            res.check_returncode()
            return res.stdout.decode("utf-8")



"""

Kubernetes workload plugin for creating infra from a multi-doc yaml file.

This is possible, but not very clean for management of resources.  The python
kubernetes api offers a helper for creating resources, but not easy approach
for removing the resources.

"""

import logging
from typing import List, Any, Dict

import yaml

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures
from mirantis.testing.metta.client import METTA_PLUGIN_INTERFACE_ROLE_CLIENT

from .kubeapi_client import KubernetesApiClientPlugin, METTA_PLUGIN_ID_KUBERNETES_CLIENT

logger = logging.getLogger("metta.contrib.kubernetes.workload.yaml")

KUBERNETES_YAML_WORKLOAD_CONFIG_LABEL = "kubernetes"
KUBERNETES_YAML_WORKLOAD_CONFIG_BASE = "workload.yaml"

KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_NAMESPACE = "namespace"
KUBERNETES_YAML_WORKLOAD_CONFIG_DEFAULT_NAMESPACE = "default"
KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_FILE = "file"
KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_YAML = "yaml"

METTA_PLUGIN_ID_KUBERNETES_YAML_WORKLOAD = "metta_kubernetes_yaml_workload"
""" workload plugin_id for the metta_kubernetes yaml plugin """


# pylint: disable=too-many-instance-attributes
class KubernetesYamlWorkloadPlugin:
    """Metta workload plugin for Kubernetes workloads created from yaml."""

    def __init__(
        self,
        environment,
        instance_id,
        label: str = KUBERNETES_YAML_WORKLOAD_CONFIG_LABEL,
        base: Any = KUBERNETES_YAML_WORKLOAD_CONFIG_BASE,
    ):
        """Run the super constructor but also set class properties.

        This implements the args part of the client interface.

        Here we expect to receive a path to a KUBECONFIG file with a context
        set and we create a Kubernetes client for use.  After that this can
        provide Core api clients as per the kubernetes SDK

        Parameters:
        -----------
        config_file (str): String path to the kubernetes config file to use

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self._config_label = label
        """ configerus load label that should contain all of the config """
        self._config_base = base
        """ configerus get key that should contain all tf config """

        workload_config = self._environment.config().load(self._config_label)

        self.namespace = workload_config.get(
            [self._config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_NAMESPACE],
            default=KUBERNETES_YAML_WORKLOAD_CONFIG_DEFAULT_NAMESPACE,
        )

        # YAML config needs to come from a yaml file path or inline yaml config
        self.resource_yaml: Dict[str, Any] = workload_config.get(
            [self._config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_YAML], default={}
        )
        """Inline YAML/Dict to source k8s resources."""
        self.file = workload_config.get(
            [self._config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_FILE], default=""
        )
        """YAML file which will source the k8s resources."""

        if self.resource_yaml is None and self.file == "":
            raise ValueError("Either inline yaml or a file path to a yaml is required.")

        self.client: KubernetesApiClientPlugin = None
        """KubeAPI client to connect to the cluster (see prepare())."""

        self.k8s_objects: List[object] = []
        """List of resource creation objects."""

        # do an initial prepare in case it is never properly run
        try:
            self.prepare()
        # pylint: disable=broad-except
        except Exception:
            pass

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Return dict data about this plugin for introspection."""
        workload_config = self._environment.config().load(self._config_label)

        return {
            "workload": {
                "deployment": {
                    "namespace": workload_config.get(
                        [
                            self._config_base,
                            KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_NAMESPACE,
                        ],
                        default=KUBERNETES_YAML_WORKLOAD_CONFIG_DEFAULT_NAMESPACE,
                    ),
                    "yaml": workload_config.get(
                        [self._config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_YAML],
                        default="None",
                    ),
                    "file": workload_config.get(
                        [self._config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_FILE],
                        default="None",
                    ),
                },
                "required_fixtures": {
                    "kubernetes": {
                        "interfaces": [METTA_PLUGIN_INTERFACE_ROLE_CLIENT],
                        "plugin_id": METTA_PLUGIN_ID_KUBERNETES_CLIENT,
                    }
                },
            }
        }

    def prepare(self, fixtures: Fixtures = None):
        """Get the kubeapi client from a set of fixtures.

        Parameters:
        -----------
        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes client plugin.

        """
        if fixtures is None:
            fixtures = self._environment.fixtures()

        try:
            self.client = fixtures.get_plugin(plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT)
        except KeyError as err:
            raise NotImplementedError(
                "Workload could not find the needed client: "
                f"{METTA_PLUGIN_ID_KUBERNETES_CLIENT}"
            ) from err

    def apply(self):
        """Use the passed yaml to create k8s resources.

        Returns:
        --------
        Any created resources.

        """
        if self.file:
            with open(self.file, encoding="utf8") as res_file:
                resources_yaml = yaml.safe_load_all(res_file)

                for resource in resources_yaml:
                    self.k8s_objects.append(
                        self.client.utils_create_from_dict(data=resource, namespace=self.namespace)
                    )

        else:
            self.k8s_objects.append(
                self.client.utils_create_from_dict(
                    data=self.resource_yaml, namespace=self.namespace
                )
            )

        return self.k8s_objects

    # @TODO write the docker container rm
    def destroy(self):
        """Destroy a created docker run."""



"""

Metta Terraform.

metta contrib functionality for terraform.  Primarily a terraform provisioner
plugin.

"""
from typing import Dict, Any

from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER
from mirantis.testing.metta.client import METTA_PLUGIN_INTERFACE_ROLE_CLIENT
from mirantis.testing.metta_cli.base import METTA_PLUGIN_INTERFACE_ROLE_CLI

from .cli import TerraformCliPlugin, METTA_TERRAFORM_CLI_PLUGIN_ID
from .provisioner import (
    TerraformProvisionerPlugin,
    TERRAFORM_PROVISIONER_CONFIG_LABEL,
    METTA_TERRAFORM_PROVISIONER_PLUGIN_ID,
)
from .client import TerraformClientPlugin, METTA_TERRAFORM_CLIENT_PLUGIN_ID


@Factory(
    plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID,
    interfaces=[
        METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER,
        METTA_TERRAFORM_PROVISIONER_PLUGIN_ID,
    ],
)
def metta_plugin_factory_provisioner_terraform(
    environment: Environment,
    instance_id: str,
    label: str = TERRAFORM_PROVISIONER_CONFIG_LABEL,
    base: Any = LOADED_KEY_ROOT,
) -> TerraformProvisionerPlugin:
    """Create an metta provisioners plugin."""
    return TerraformProvisionerPlugin(environment, instance_id, label, base)


@Factory(
    plugin_id=METTA_TERRAFORM_CLIENT_PLUGIN_ID,
    interfaces=[
        METTA_PLUGIN_INTERFACE_ROLE_CLIENT,
        METTA_TERRAFORM_CLIENT_PLUGIN_ID,
    ],
)
# pylint: disable=too-many-arguments
def metta_plugin_factory_client_terraform(
    environment: Environment,
    instance_id: str,
    chart_path: str,
    state_path: str,
    tfvars: Dict[str, Any],
    tfvars_path: str,
) -> TerraformClientPlugin:
    """Create an metta client plugin."""
    return TerraformClientPlugin(
        environment,
        instance_id,
        chart_path=chart_path,
        state_path=state_path,
        tfvars=tfvars,
        tfvars_path=tfvars_path,
    )


@Factory(
    plugin_id=METTA_TERRAFORM_CLI_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI],
)
def metta_terraform_factory_cli_terraform(
    environment: Environment, instance_id: str
) -> TerraformCliPlugin:
    """Create a cli plugin."""
    return TerraformCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint METTA BootStrapping -----


# pylint: disable=unused-argument
def bootstrap_environment(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added.

    """



"""

Structures for interpreting Sonobuoy results.

"""
import os
import json
import logging
from enum import Enum, unique
from typing import List, Dict, Any
import subprocess

import yaml

logger = logging.getLogger("sonobuoy:results")


@unique
# pylint: disable=too-few-public-methods
class Status(Enum):
    """Enumerator to plugin states."""

    ERROR = "error"
    """ an error occured trying to interpret sonobuoy """
    PENDING = "pending"
    """ still pending """
    RUNNING = "running"
    """ testing is running """
    FAILED = "failed"
    """ testing has failed """
    COMPLETE = "complete"
    """ testing has completed without failure """
    PASSED = "passed"
    """ testing has passed """
    POSTPROCESS = "post-processing"
    """ testing has finished and is being processed """


class SonobuoyStatus:
    """A status output from the sonobuoy CLI."""

    def __init__(self, status_json: str):
        """Build from sonobuoy status results."""
        try:
            status = json.loads(status_json)
            self.status: Status = Status(status["status"])
            self.tar_info: str = status["tar-info"]

            self.plugins: Dict[str, Dict[str, Any]] = {}
            for plugin in status["plugins"]:
                self.plugins[plugin["plugin"]] = plugin

        except json.decoder.JSONDecodeError:
            logger.warning("json decoding of status failed: %s", status_json)
            self.status = Status.ERROR
            self.plugins = {}

    def plugin_list(self):
        """Retrieve the list of plugins."""
        return list(self.plugins.keys())

    def plugin(self, plugin: str):
        """Retrieve the results for one plugin."""
        return self.plugins[plugin]

    def plugin_status(self, plugin: str) -> "Status":
        """Get the status code for a plugin."""
        status_string = self.plugin(plugin)["status"]
        return Status(status_string)

    def __str__(self) -> str:
        """Convert to a string."""
        status: List[str] = []
        for plugin_id in self.plugin_list():
            status.append(f"{plugin_id}:{self.plugin_status(plugin_id)}")
        return f"[{']['.join(status)}]"


class SonobuoyResultsPluginItem:
    """An individual item from a sonobuoy results plugin."""

    def __init__(self, item_dict: Dict[str, Any]):
        """Single plugin result item."""
        self.name = item_dict["name"]
        self.status = Status(item_dict["status"])
        self.meta = item_dict["meta"]
        self.details = item_dict["details"] if "details" in item_dict else {}

    def meta_file_path(self):
        """Get the path to the error item file."""
        return self.meta["file"]

    def meta_file(self):
        """Get the contents of the file."""
        with open(self.meta_file_path(), encoding="utf8") as meta_file:
            return yaml.safe_load(meta_file)


class SonobuoyResultsPlugin:
    """The full results for a plugin."""

    def __init__(self, path: str):
        """Load results for a plugin results call."""
        with open(os.path.join(path, "sonobuoy_results.yaml"), encoding="utf8") as results_yaml:
            self.summary = yaml.safe_load(results_yaml)

    def name(self) -> str:
        """Return string name of plugin."""
        return self.summary["name"]

    def status(self) -> "Status":
        """Return the status object for the plugin results."""
        return Status(self.summary["status"])

    def __len__(self):
        """Count how many items are in the plugin_results."""
        return len(self.summary["items"])

    def __getitem__(self, instance_id: Any) -> SonobuoyResultsPluginItem:
        """Get item details from the plugin results."""
        return SonobuoyResultsPluginItem(item_dict=self.summary["items"][instance_id])


class SonobuoyResults:
    """Results retrieved analyzer."""

    def __init__(self, tarball: str, folder: str):
        """Interpret tarball contents."""
        logger.debug("un-tarring retrieved results: %s", tarball)
        res = subprocess.run(["tar", "-xzf", tarball, "-C", folder], check=True, text=True)
        res.check_returncode()

        self.results_path = folder

        with open(os.path.join(folder, "meta", "config.json"), encoding="utf8") as config_json:
            self.meta_config = json.load(config_json)
        with open(os.path.join(folder, "meta", "info.json"), encoding="utf8") as info_json:
            self.meta_info = json.load(info_json)
        with open(os.path.join(folder, "meta", "query-time.json"), encoding="utf8") as qt_json:
            self.meta_querytime = json.load(qt_json)

        self.plugins = []
        for plugin_id in self.meta_info["plugins"]:
            self.plugins.append(plugin_id)

    def plugin_list(self):
        """Return a string list of plugin ids."""
        return self.plugins

    def plugin(self, plugin_id) -> SonobuoyResultsPlugin:
        """Return the results for a single plugin."""
        return SonobuoyResultsPlugin(os.path.join(self.results_path, "plugins", plugin_id))



"""

Metta CLI : Ansible Provisioner commands.

Comamnds for itneracting with the ansible plugins, primarily the provisioner

"""
import logging

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta_health.healthcheck import Health

from mirantis.testing.metta_cli.base import CliBase, cli_output

from .ansiblecli_provisioner import METTA_ANSIBLE_ANSIBLECLIPLAYBOOK_PROVISIONER_PLUGIN_ID
from .ansiblecli_client import (
    METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID,
    METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKCLIENT_PLUGIN_ID,
)
from .ansiblecli_workload import (
    METTA_ANSIBLE_ANSIBLECLI_COREWORKLOAD_PLUGIN_ID,
    METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKWORKLOAD_PLUGIN_ID,
)

logger = logging.getLogger("metta.cli.ansible")

METTA_ANSIBLE_CLI_PLUGIN_ID = "metta_ansible_cli"
""" cli plugin_id for the info plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class AnsibleCliPlugin(CliBase):
    """Fire command/group generator for Ansible commands."""

    def fire(self):
        """Return a dict of commands."""
        commands = {}

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            commands["ansible"] = AnsibleCliClientGroup(self._environment)

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKCLIENT_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            commands["ansible-playbook"] = AnsibleCliPlaybookClientGroup(self._environment)

        return commands


class AnsibleCliClientGroup:
    """Commands for interacting with a metta ansible core client plugin."""

    def __init__(self, environment: Environment):
        """Inject environment into command gorup."""
        self._environment: Environment = environment

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_ANSIBLE_ANSIBLECLI_COREWORKLOAD_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            self.workload = AnsibleCliWorkgroupGroup(self._environment)

    def _select_client(self, instance_id: str = ""):
        """Pick a matching client."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=[METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID],
                instance_id=instance_id,
            )

        # Get the highest priority provisioner
        return self._environment.fixtures().get(
            plugin_id=METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID,
        )

    def info(self, client: str = "", deep: bool = False):
        """Get info about a client plugin."""
        fixture = self._select_client(instance_id=client)
        return cli_output(fixture.info(deep=deep))

    def debug(self, client: str = "", hosts: str = "all"):
        """Run client plugin debug module."""
        fixture = self._select_client(instance_id=client)
        plugin = fixture.plugin
        return cli_output(plugin.debug(hosts=hosts))

    def setup(self, client: str = "", hosts: str = "all"):
        """Run client plugin setup module."""
        fixture = self._select_client(instance_id=client)
        plugin = fixture.plugin
        return cli_output(plugin.setup(hosts=hosts))

    def ping(self, client: str = "", hosts: str = "all"):
        """Run client plugin ping module."""
        fixture = self._select_client(instance_id=client)
        plugin = fixture.plugin
        return cli_output(plugin.ping(hosts=hosts))

    def health(self, client: str = ""):
        """Run client healthcheck."""
        fixture = self._select_client(instance_id=client)
        plugin = fixture.plugin
        health: Health = plugin.health()
        return cli_output(
            {
                "fixture": {
                    "plugin_id": fixture.plugin_id,
                    "instance_id": fixture.instance_id,
                    "priority": fixture.priority,
                },
                "status": health.status(),
                "messages": list(health.messages()),
            }
        )


class AnsibleCliPlaybookClientGroup:
    """Commands for interacting with a metta ansible playbook client plugin."""

    def __init__(self, environment: Environment):
        """Inject environment into command gorup."""
        self._environment: Environment = environment

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_ANSIBLE_ANSIBLECLIPLAYBOOK_PROVISIONER_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            self.provisioner = AnsibleCliPlaybookProvisionerGroup(self._environment)

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKWORKLOAD_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            self.workload = AnsibleCliPlaybookWorkgroupGroup(self._environment)

    def _select_client(self, instance_id: str = ""):
        """Pick a matching client."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=[METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID],
                instance_id=instance_id,
            )

        # Get the highest priority provisioner
        return self._environment.fixtures().get(
            plugin_id=METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID,
        )

    def info(self, client: str = "", deep: bool = False):
        """Get info about a client plugin."""
        fixture = self._select_client(instance_id=client)
        return cli_output(fixture.info(deep=deep))

    def run(self, playbook: str, client: str = ""):
        """Get info about a client plugin."""
        fixture = self._select_client(instance_id=client)
        plugin = fixture.plugin
        return cli_output(plugin.run_file(playbook))


class AnsibleCliPlaybookProvisionerGroup:
    """Commands for interacting with a metta ansible provisioner plugin."""

    def __init__(self, environment: Environment):
        """Inject environment into command gorup."""
        self._environment: Environment = environment

    def _select_provisioner(self, instance_id: str = ""):
        """Pick a matching provisioner."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=[METTA_ANSIBLE_ANSIBLECLIPLAYBOOK_PROVISIONER_PLUGIN_ID],
                instance_id=instance_id,
            )

        # Get the highest priority provisioner
        return self._environment.fixtures().get(
            plugin_id=METTA_ANSIBLE_ANSIBLECLIPLAYBOOK_PROVISIONER_PLUGIN_ID,
        )

    def info(self, provisioner: str = "", deep: bool = False):
        """Get info about a provisioner plugin."""
        fixture = self._select_provisioner(instance_id=provisioner)
        return cli_output(fixture.info(deep=deep))

    def prepare(self, provisioner: str = ""):
        """Run provisioner prepare."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.prepare(self._environment.fixtures())

    def apply(self, provisioner: str = ""):
        """Run provisioner apply."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.apply()

    def destroy(self, provisioner: str = ""):
        """Run provisioner destroy."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.destroy()


class AnsibleCliWorkgroupGroup:
    """Commands for interacting with a metta ansible workgroup plugin."""

    def __init__(self, environment: Environment):
        """Inject environment into command gorup."""
        self._environment: Environment = environment

    def _select_workgroup(self, instance_id: str = ""):
        """Pick a matching workgroup."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=[METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID],
                instance_id=instance_id,
            )

        # Get the highest priority provisioner
        return self._environment.fixtures().get(
            plugin_id=METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID,
        )

    def info(self, workgroup: str = "", deep: bool = False):
        """Get info about the plugin."""
        fixture = self._select_workgroup(instance_id=workgroup)
        return cli_output(fixture.info(deep=deep))

    def apply(self, workgroup: str = ""):
        """Apply the plygin workload."""
        fixture = self._select_workgroup(instance_id=workgroup)
        plugin = fixture.plugin
        plugin.prepare(self._environment.fixtures())
        plugin.apply()


class AnsibleCliPlaybookWorkgroupGroup:
    """Commands for interacting with a metta ansible playbook workgroup plugin."""

    def __init__(self, environment: Environment):
        """Inject environment into command gorup."""
        self._environment: Environment = environment

    def _select_workgroup(self, instance_id: str = ""):
        """Pick a matching workgroup."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=[METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKWORKLOAD_PLUGIN_ID],
                instance_id=instance_id,
            )

        # Get the highest priority provisioner
        return self._environment.fixtures().get(
            plugin_id=METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKWORKLOAD_PLUGIN_ID,
        )

    def info(self, workgroup: str = "", deep: bool = False):
        """Get info about the plugin."""
        fixture = self._select_workgroup(instance_id=workgroup)
        fixture.plugin.prepare()
        return cli_output(fixture.info(deep=deep))

    def apply(self, workgroup: str = ""):
        """Apply the plugin workload."""
        fixture = self._select_workgroup(instance_id=workgroup)
        plugin = fixture.plugin
        plugin.prepare(self._environment.fixtures())
        plugin.apply()



"""

Launchpad metta provisioner plugin.

Provisioner/Install Mirantis products onto an existing cluster using a
Launchpad client plugin.  This is effectively a provisioner style wrapper
for the client plugin, which this plugin will create from provisioner
configuration.

"""
import os.path
import logging
from typing import Any, List, Dict
import yaml

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import (
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL,
)
from configerus.validator import ValidationError

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures

from .client import (
    METTA_LAUNCHPAD_CLIENT_PLUGIN_ID,
    LaunchpadClientPlugin,
)
from .launchpad import (
    METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT,
    METTA_LAUNCHPADCLIENT_WORKING_DIR_DEFAULT,
)

logger = logging.getLogger("metta_launchpad:provisioner")

METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID = "metta_launchpad_provisioner"
""" Metta plugin_id for the launchpad provisioner plugin """

METTA_LAUNCHPAD_CONFIG_LABEL = "launchpad"
"""Launchpad config label for configuration"""
METTA_LAUNCHPAD_CLIENT_SYSTEMS_KEY = "systems"
"""If provided, this config key provide a dictionary of configuration of client system plugins."""
METTA_LAUNCHPAD_CONFIG_ROOT_PATH_KEY = "root.path"
"""Config key for a base file path that should be used for any relative paths"""
METTA_LAUNCHPAD_CONFIG_KEY = "config"
"""Which config key will provide the launchpad yml"""
METTA_LAUNCHPAD_CLI_CONFIG_FILE_KEY = "config_file"
"""Launchpad config cli key to tell us where to put the launchpad yml file"""
METTA_LAUNCHPAD_CLI_WORKING_DIR_KEY = "working_dir"
"""Launchpad config cli configuration working dir key """
METTA_LAUNCHPAD_CLI_CLUSTEROVERRIDE_KEY = "cluster_name"
"""If provided, this config key will override a cluster name pulled from yaml"""
METTA_LAUNCHPAD_CLI_OPTIONS_KEY = "cli"
"""If provided, these will be passed to the launchpad client to be used on all operations"""

METTA_LAUNCHPAD_VALIDATE_JSONSCHEMA = {
    "type": "object",
    "properties": {
        "config": {
            "type": ["object", "null"],
        },
        "systems": {
            "type": "object",
            "properties": {
                "mirantis_mke_client": {
                    "type": "object",
                    "properties": {
                        "accesspoint": {"type": ["string", "null"]},
                        "username": {"type": ["string", "null"]},
                        "password": {"type": ["string", "null"]},
                        "client_bundle_root": {"type": ["string", "null"]},
                    },
                },
                "mirantis_msr_client": {
                    "type": "object",
                    "properties": {
                        "accesspoint": {"type": ["string", "null"]},
                        "username": {"type": ["string", "null"]},
                        "password": {"type": ["string", "null"]},
                    },
                },
            },
        },
        "cli": {
            "accept-license": {"type": "bool"},
            "disable-telemetry": {"type": "bool"},
        },
        "config_file": {"type": "string"},
        "working_dir": {"type": "string"},
    },
    "required": ["config_file"],
}
""" Validation jsonschema for Launchpad config contents """
METTA_LAUNCHPAD_PROVISIONER_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: METTA_LAUNCHPAD_VALIDATE_JSONSCHEMA
}
""" configerus validation target to match validate Launchpad config """

METTA_LAUNCHPAD_CONFIG_VALIDATE_JSONSCHEMA = {
    "type": "object",
    "properties": {
        "apiVersion": {"type": "string"},
        "kind": {"type": "string"},
        "hosts": {"type": "array", "items": {"type": "object"}},
        "spec": {
            "type": "object",
            "properties": {
                "mcr": {"type": "object"},
                "mke": {"type": "object"},
                "msr": {"type": "object"},
            },
            "required": ["mcr", "mke"],
        },
    },
    "required": ["apiVersion", "kind", "spec"],
}
""" Validation jsonschema for launchpad configuration """
METTA_LAUNCHPAD_CONFIG_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: METTA_LAUNCHPAD_CONFIG_VALIDATE_JSONSCHEMA
}
""" configerus jsonschema validation target for launchpad config """


# pylint: disable=too-many-instance-attributes
class LaunchpadProvisionerPlugin:
    """Launchpad provisioner class.

    Use this to provision a system using Mirantis launchpad

    """

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        label: str = METTA_LAUNCHPAD_CONFIG_LABEL,
        base: Any = LOADED_KEY_ROOT,
    ):
        """Configure a new Launchpad provisioner plugin instance."""
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self._config_label = label
        """ configerus load label that should contain all of the config """
        self._config_base = base
        """ configerus get key that should contain all tf config """

        self.fixtures = Fixtures()
        """ keep a collection of fixtures that this provisioner creates """

        # attempt to be declarative and make the client plugin in case the
        # terraform chart has already been run.
        try:
            # Make the child client plugin.
            self.make_fixtures()

        # dont' block the construction on an exception
        # pylint: disable=broad-except
        except Exception:
            pass

    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Get info about the plugin.

        Returns:
        --------
        Dict of introspective information about this plugin_info

        """
        # Loaded plugin configuration
        launchpad_config_loaded = self._environment.config().load(self._config_label)

        return {
            "plugin": {
                "config_label": self._config_label,
                "config_base": self._config_base,
            },
            "config": {
                "working_dir": launchpad_config_loaded.get(
                    [self._config_base, METTA_LAUNCHPAD_CLI_WORKING_DIR_KEY], default="MISSING"
                ),
                "root_path": launchpad_config_loaded.get(
                    [self._config_base, METTA_LAUNCHPAD_CONFIG_ROOT_PATH_KEY], default="NONE"
                ),
                "config_file": launchpad_config_loaded.get(
                    [self._config_base, METTA_LAUNCHPAD_CLI_CONFIG_FILE_KEY], default="MISSING"
                ),
                "cli_options": launchpad_config_loaded.get(
                    [self._config_base, METTA_LAUNCHPAD_CLI_OPTIONS_KEY], default="NONE"
                ),
            },
            "client": self._get_client_plugin().info(deep=deep),
        }

    # pylint: disable=no-self-use
    def prepare(self):
        """Prepare the provisioning cluster for install.

        We ignore this.

        """
        logger.info("Running Launchpad Prepare().  Launchpad has no prepare stage.")

    def apply(self, debug: bool = False):
        """Bring a cluster up.

        Not that we re-write the yaml file as it may depend on config which was not
        available when this object was first constructed.

        Raises:
        -------
        ValueError if the object has been configured (prepare) with config that
            doesn't work, or if the backend doesn't give valid yml

        Exception if launchpad fails.

        """
        logger.info("Using launchpad to install products onto backend cluster")
        self._write_launchpad_yml()
        self.make_fixtures()  # we wouldn't need this if we could update the systems
        self._get_client_plugin().apply(debug=debug)

    def destroy(self):
        """Ask the client to remove installed resources."""
        if self._has_launchpad_yml():
            logger.info("Using launchpad to remove installed products from the backend cluster")
            self._get_client_plugin().reset()
            self._rm_launchpad_yml()

    # ----- CLUSTER INTERACTION -----

    def _has_launchpad_yml(self) -> bool:
        """Check if the launchpad yml file exists."""
        # Loaded configerus config for the plugin. Ready for .get().
        plugin_config = self._environment.config().load(self._config_label)

        config_file: str = plugin_config.get(
            [self._config_base, METTA_LAUNCHPAD_CLI_CONFIG_FILE_KEY],
            default=METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT,
        )
        return bool(config_file) and os.path.exists(config_file)

    def _write_launchpad_yml(self):
        """Write config contents to a yaml file for launchpad."""
        self._rm_launchpad_yml()

        # load and validation all of the launchpad configuration.
        launchpad_loaded = self._environment.config().load(
            self._config_label,
            validator=METTA_LAUNCHPAD_PROVISIONER_VALIDATE_TARGET,
            force_reload=True,
        )

        # load all of the launchpad configuration, force a reload to get up to date contents
        config_contents: Dict[str, Any] = launchpad_loaded.get(
            [self._config_base, METTA_LAUNCHPAD_CONFIG_KEY],
            validator=METTA_LAUNCHPAD_CONFIG_VALIDATE_TARGET,
        )

        # decide on a path for the runtime launchpad.yml file
        config_path: str = os.path.realpath(
            launchpad_loaded.get(
                [self._config_base, METTA_LAUNCHPAD_CLI_CONFIG_FILE_KEY],
                default=METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT,
            )
        )

        # Our launchpad config differs slightly from the schema that launchpad
        # consumes, so we need a small conversion
        config_contents = self._convert_launchpad_config_to_file_format(config_contents)

        # write the launchpad output to our yaml file target (after creating the path)
        logger.debug(
            "Updating launchpad yaml file: %s =>/n%s",
            config_path,
            yaml.dump(config_contents),
        )
        with open(config_path, "w", encoding="utf8") as config_file_object:
            yaml.dump(config_contents, config_file_object)

    def _rm_launchpad_yml(self):
        """Update config and write the cfg and inventory files."""
        # Loaded configerus config for the plugin. Ready for .get().
        plugin_config = self._environment.config().load(self._config_label)

        config_file: str = plugin_config.get(
            [self._config_base, METTA_LAUNCHPAD_CLI_CONFIG_FILE_KEY],
            default=METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT,
        )
        if config_file and os.path.exists(config_file):
            logger.debug("Launchpad provisioner removing created files.")
            os.remove(config_file)

    def make_fixtures(self):
        """Make the client plugin for terraform interaction."""
        try:
            # load and validation all of the launchpad configuration.
            launchpad_config_loaded = self._environment.config().load(
                self._config_label,
                validator=METTA_LAUNCHPAD_PROVISIONER_VALIDATE_TARGET,
                force_reload=True,
            )
        except ValidationError as err:
            raise ValueError("Launchpad config failed validation.") from err

        # if launchpad needs to be run in a certain path, set it with this config
        working_dir: str = launchpad_config_loaded.get(
            [self._config_base, METTA_LAUNCHPAD_CLI_WORKING_DIR_KEY],
            default=METTA_LAUNCHPADCLIENT_WORKING_DIR_DEFAULT,
        )

        # decide on a path for the runtime launchpad.yml file
        config_file: str = launchpad_config_loaded.get(
            [self._config_base, METTA_LAUNCHPAD_CLI_CONFIG_FILE_KEY],
            default=METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT,
        )
        # List of launchpad cli options to pass to the client for all operations.
        cli_options: Dict[str, Any] = launchpad_config_loaded.get(
            [self._config_base, METTA_LAUNCHPAD_CLI_OPTIONS_KEY], default={}
        )
        # List of systems that the client should configure for children plugins.
        systems: Dict[str, Dict[str, Any]] = launchpad_config_loaded.get(
            [self._config_base, METTA_LAUNCHPAD_CLIENT_SYSTEMS_KEY], default={}
        )

        fixture = self._environment.new_fixture(
            plugin_id=METTA_LAUNCHPAD_CLIENT_PLUGIN_ID,
            instance_id=self.client_instance_id(),
            priority=70,
            arguments={
                "config_file": config_file,
                "working_dir": working_dir,
                "cli_options": cli_options,
                "systems": systems,
            },
            labels={
                "parent_plugin_id": METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID,
                "parent_instance_id": self._instance_id,
            },
            replace_existing=True,
        )
        # keep this fixture attached to the workload to make it retrievable.
        self.fixtures.add(fixture, replace_existing=True)

    def client_instance_id(self) -> str:
        """Construct an instanceid for the child client plugin."""
        return f"{self._instance_id}-{METTA_LAUNCHPAD_CLIENT_PLUGIN_ID}"

    def _get_client_plugin(self) -> LaunchpadClientPlugin:
        """Retrieve the client plugin if we can."""
        try:
            return self.fixtures.get_plugin(plugin_id=METTA_LAUNCHPAD_CLIENT_PLUGIN_ID)
        except KeyError as err:
            raise RuntimeError(
                "Launchpad provisioner cannot find its client plugin, and "
                "cannot process any client actions.  Was a client created?"
            ) from err

    def _convert_launchpad_config_to_file_format(self, config):
        """Convert our launchpad config to the schema that launchpad uses."""
        # 1 discover the hosts counts
        hosts = []
        managers = []
        workers = []
        msrs = []
        for host in config["spec"]["hosts"]:
            hosts.append(host)
            if host["role"] == "manager":
                managers.append(host)
            if host["role"] == "worker":
                workers.append(host)
            if host["role"] == "msr":
                msrs.append(host)

        # convert install flags and update flags to lists from dicts
        def dtol(dic):
            """Convert dict flags to lists."""
            items: List[str] = []
            for key, value in dic.items():
                if value is True:
                    items.append(f"--{key}")
                else:
                    items.append(f"--{key}={value}")
            return items

        try:
            config["spec"]["mke"]["installFlags"] = dtol(config["spec"]["mke"]["installFlags"])
        except KeyError:
            pass
        try:
            config["spec"]["mke"]["upgradeFlags"] = dtol(config["spec"]["mke"]["upgradeFlags"])
        except KeyError:
            pass
        try:
            config["spec"]["msr"]["installFlags"] = dtol(config["spec"]["msr"]["installFlags"])
        except KeyError:
            pass
        try:
            config["spec"]["msr"]["upgradeFlags"] = dtol(config["spec"]["msr"]["upgradeFlags"])
        except KeyError:
            pass

        # If no msrs, then drop the msr block and force the type.
        if len(msrs) == 0:
            config["kind"] = "mke"
            config["spec"].pop("msr")

        return config



"""

Dummy workload plugin.

Dummy workload plugin that really only holds fixtures of its own.  Used for
mocking and testing.

"""
from typing import Dict, Any
import logging

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures

logger = logging.getLogger("metta.contrib.dummy.workload")


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class DummyWorkloadPlugin:
    """Dummy workload class."""

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        fixtures: Dict[str, Dict[str, Any]] = None,
    ):
        """Set class properties.

        Parameters:
        -----------
        outputs (Dict[Dict]) : pass in a dictionary which defines outputs that
            should be returned

        clients (Dict[Dict]) : pass in a dictionary which defines which clients
            should be requested when working on a provisioner

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self.fixtures = Fixtures()
        """This plugin keeps fixtures."""
        if fixtures is not None:
            for child_instance_id, child_instance_dict in fixtures.items():
                child = environment.add_fixture_from_dict(
                    instance_id=child_instance_id, plugin_dict=child_instance_dict
                )
                self.fixtures.add(child)



"""

Metta Output plugin that holds structured Dict set_data.

The data can be navigated as a configerus Loaded data set.
Assigned data can also be given a configerus validator topic
which will enforce validation on the content, and raise
an exception on invalid assignment.

"""
import logging
from typing import Dict

from configerus.loaded import Loaded, LOADED_KEY_ROOT

from mirantis.testing.metta.environment import Environment

logger = logging.getLogger("metta.contrib.common.output.dict")

METTA_PLUGIN_ID_OUTPUT_DICT = "dict"
""" output plugin_id for the dict plugin """


class DictOutputPlugin:
    """Metta Output plugin for a Dict output type.

    This output plugin leverages configurators features, treating the dict as
    a config source, in order to get validation and navigations.

    """

    def __init__(
        self, environment: Environment, instance_id: str, data: Dict = None, validator: str = ""
    ):
        """Run the super constructor but also set class properties.

        Here we treat the data dict as a configerus.loaded.Loaded instance,
        with out config as a parent, so that we can leverage the configerus
        tools for searching, formating and validation.

        a configerus.loaded.Loaded object is kept for the config.

        Parameters:
        -----------
        data (Dict) : any dict data to be stored

        validator (str) : a configerus validator target if you want valdiation
            applied to the data before it is added.

        Raises:
        -------
        A configerus.validate.ValidationError is raised if you passed in a
        validator target and validation failed.

        An AssertionError is raised if you didn't pass in a Dict.

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        if data is None:
            data = {}

        self.set_data(data, validator)

    def set_data(self, data: Dict, validator: str = ""):
        """Re-set the data for the output."""
        assert isinstance(data, dict), f"Expected Dict of data, got {data}"

        if validator:
            self._environment.config().validate(data, validator)

        mock_instance_id = f"dict-output-{self._instance_id}"
        self.loaded = Loaded(
            data=data, parent=self._environment.config(), instance_id=mock_instance_id
        )

    def get_output(self, key: str = LOADED_KEY_ROOT, validator: str = ""):
        """Retrieve an output.

        Because we treated that data as a high-priority configerus source with
        a custom label, we can retrieve data from that source easily and also
        leverage other configerus options such as templating and validation

        If you don't pass a key, you will get the entire data value

        Parameters:
        -----------
        key (str) : you can optionally pass a key to retrieve only a part of the
            data structure.  This uses the configerus .get() command which uses
            dot "." notation to descend a tree.

        validator (str) : you can tell configerus to apply a validator to the
            return value

        Returns:
        --------
        Any of retreived data in the assigned data, as though you were making
        a configerus.loaded.Loaded.get()

        Raises:
        -------
        AttributeError if you are trying to get output before you have
        assigned data using .arguments()

        configerus.validate.ValidationError if you passed in a validator and
        validation failed.

        """
        return self.loaded.get(key, validator=validator)

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Return dict data about this plugin for introspection."""
        return {"output": {"data": self.loaded.data}}



"""

Run a Sonobuoy run on a k82 client.

Use this to run the sonobuoy implementation

"""
from typing import Any, List, Dict
import logging

import yaml

from configerus.loaded import Loaded, LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import (
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL,
)
from configerus.validator import ValidationError

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures
from mirantis.testing.metta.client import METTA_PLUGIN_INTERFACE_ROLE_CLIENT
from mirantis.testing.metta_kubernetes import (
    METTA_PLUGIN_ID_KUBERNETES_CLIENT,
    KubernetesApiClientPlugin,
)

from .client import METTA_SONOBUOY_CLIENT_PLUGIN_ID, SonobuoyClientPlugin
from .plugin import Plugin
from .sonobuoy import (
    SONOBUOY_DEFAULT_RESULTS_PATH,
)
from .results import (
    SonobuoyStatus,
    SonobuoyResults,
)

logger = logging.getLogger("workload.sonobuoy")

METTA_SONOBUOY_WORKLOAD_PLUGIN_ID = "metta_sonobuoy_workload"
""" workload plugin_id for the sonobuoy plugin """

SONOBUOY_WORKLOAD_CONFIG_LABEL = "sonobuoy"
""" Configerus label for retrieving sonobuoy config """
SONOBUOY_WORKLOAD_CONFIG_BASE = LOADED_KEY_ROOT
""" Configerus get base for retrieving the default workload config """

SONOBUOY_CONFIG_KEY_MODE = "mode"
""" config key for mode """
SONOBUOY_CONFIG_KEY_KUBERNETESVERSION = "kubernetes.version"
""" config key for kubernetes version """
SONOBUOY_CONFIG_KEY_PLUGINS = "plugins"
""" config key for what plugins to run """
SONOBUOY_CONFIG_KEY_PLUGINDEF = "definition"
""" config key for plugin definition """
SONOBUOY_CONFIG_KEY_PLUGINPATH = "path"
""" config key for plugin file path """
SONOBUOY_CONFIG_KEY_PLUGINENVS = "envs"
""" config key for plugin env vars"""
SONOBUOY_CONFIG_KEY_RESULTSPATH = "results.path"
""" config key for path to put results """

# THIS IS OUT OF DATSE
SONOBUOY_VALIDATE_JSONSCHEMA = {
    "type": "object",
    "properties": {"plugins": {"$ref": "#/definitions/plugin"}},
    "definitions": {
        "plugin": {
            "type": "object",
            "properties": {
                # Optionally provide a plugin path
                "path": {"type": "string"},
                # Optionally provide a definition
                "definition": {"type": "object"},
                # dict of plugin env variables.
                "envs": {"$ref": "string"},
            },
        }
    },
}
""" Validation jsonschema for terraform config contents """
SONOBUOY_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: SONOBUOY_VALIDATE_JSONSCHEMA
}
""" configerus validation target to match the jsonschema config """


class SonobuoyWorkloadPlugin:
    """Workload class for the Sonobuoy."""

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        label: str = SONOBUOY_WORKLOAD_CONFIG_LABEL,
        base: Any = SONOBUOY_WORKLOAD_CONFIG_BASE,
    ):
        """Initialize workload plugin.

        Parameters:
        -----------
        label (str) : Configerus label for loading config
        base (Any) : configerus base key which should contain all of the config

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        logger.info("Preparing sonobuoy settings")

        self._config_label: str = label
        """ configerus load label that should contain all of the config """
        self._config_base: str = base
        """ configerus get key that should contain all tf config """

        self.fixtures: Fixtures = Fixtures()
        """This plugin creates fixtures, so they are tracked here."""

        # go for early declarative testing to the plugin.
        try:
            self.prepare()
        # pylint: disable=broad-except
        except Exception:
            logger.debug("not able to early prepare sonobuoy.")

    # the deep argument is a standard for the info hook
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Return dict data about this plugin for introspection."""
        # get a configerus LoadedConfig for the sonobuoy label
        loaded: Loaded = self._environment.config().load(self._config_label)
        # load the sonobuoy conifg (e.g. sonobuoy.yml)
        sonobuoy_config: Dict[str, Any] = loaded.get(self._config_base, default={})

        return {
            "config": {"label": self._config_label, "base": self._config_base},
            "workload": {
                "config": sonobuoy_config,
                "required_fixtures": {
                    "kubernetes": {
                        "interface": [METTA_PLUGIN_INTERFACE_ROLE_CLIENT],
                        "plugin_id": METTA_PLUGIN_ID_KUBERNETES_CLIENT,
                    }
                },
                "run": {"args": loaded.get([self._config_base, "run"], default="NONE")},
            },
        }

    def prepare(self, fixtures: Fixtures = None):
        """Create a workload instance from a set of fixtures.

        Parameters:
        -----------
        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes api client plugin.

        """
        if fixtures is None:
            fixtures = self._environment.fixtures()

        # Retrieve and Validate the config overall using jsonschema
        try:
            # get a configerus LoadedConfig for the sonobuoy label
            loaded = self._environment.config().load(
                self._config_label, validator=SONOBUOY_VALIDATE_TARGET
            )
        except ValidationError as err:
            raise ValueError("Invalid sonobuoy config received") from err

        # We need to discover all of the plugins to run.
        #
        # For plugins with inline definitions, we need to create file definitions
        # to pass to sonobuoy.
        resources_path: str = loaded.get([self._config_base, "resources.path"], default="./")
        resources_prefix: str = loaded.get(
            [self._config_base, "resources.prefix"], default="sonobuoy-plugin-"
        )

        plugins: List[Plugin] = []
        for plugin_id in loaded.get(
            [self._config_base, SONOBUOY_CONFIG_KEY_PLUGINS], default={}
        ).keys():

            plugin_envs = loaded.get(
                [
                    self._config_base,
                    SONOBUOY_CONFIG_KEY_PLUGINS,
                    plugin_id,
                    SONOBUOY_CONFIG_KEY_PLUGINENVS,
                ],
                default=plugin_id,
            )

            # plugin_def gives us a plugin definition which defines what we pass
            # to sonobuoy using the -p flag.
            #
            # If a plugin def is missing then plugin_id is used.
            #
            # It can be one of three types:
            # 1. a core plugin id like 'e2e'
            # 2. a path to a plugin yml file which defines a plugin.
            # 3. an object which defines the plugin conf which will be written
            #    to a yaml file.
            plugin_def = loaded.get(
                [
                    self._config_base,
                    SONOBUOY_CONFIG_KEY_PLUGINS,
                    plugin_id,
                    SONOBUOY_CONFIG_KEY_PLUGINDEF,
                ],
                default="",
            )
            plugin_path = loaded.get(
                [
                    self._config_base,
                    SONOBUOY_CONFIG_KEY_PLUGINS,
                    plugin_id,
                    SONOBUOY_CONFIG_KEY_PLUGINPATH,
                ],
                default="",
            )

            if plugin_def:
                # here we received a plugin definition which we must write to
                # a file.
                if not plugin_path:
                    plugin_path = resources_path + resources_prefix + plugin_id + ".yml"

                with open(plugin_path, "w") as plugin_file:
                    yaml.dump(plugin_def, plugin_file, encoding="utf-8")
                plugin_def = plugin_path

                plugins.append(
                    Plugin(plugin_id=plugin_id, plugin_def=plugin_path, envs=plugin_envs)
                )
                continue

            if plugin_path:
                plugins.append(
                    Plugin(plugin_id=plugin_id, plugin_def=plugin_path, envs=plugin_envs)
                )
                continue

            plugins.append(Plugin(plugin_id=plugin_id, plugin_def=plugin_id, envs=plugin_envs))

        # String path to where to keep the results.
        # maybe get this from config?
        results_path: str = loaded.get(
            [self._config_base, SONOBUOY_CONFIG_KEY_RESULTSPATH],
            default=SONOBUOY_DEFAULT_RESULTS_PATH,
        )

        kubeclient: KubernetesApiClientPlugin = fixtures.get_plugin(
            plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT,
        )

        client_fixture = self._environment.new_fixture(
            plugin_id=METTA_SONOBUOY_CLIENT_PLUGIN_ID,
            instance_id=self.client_instance_id(),
            priority=70,
            arguments={
                "kubeclient": kubeclient,
                "plugins": plugins,
                "results_path": results_path,
            },
            labels={
                "container": "plugin",
                "environment": self._environment.instance_id(),
                "parent_plugin_id": METTA_SONOBUOY_WORKLOAD_PLUGIN_ID,
                "parent_instance_id": self._instance_id,
            },
            replace_existing=True,
        )
        # keep this fixture attached to the workload to make it retrievable.
        self.fixtures.add(client_fixture, replace_existing=True)

    # These are work for this scenario
    # pylint: disable=arguments-differ
    def apply(self, wait: bool = True):
        """Run sonobuoy."""
        logger.info("Starting Sonobuoy run")

        # Retrieve and Validate the config overall using jsonschema
        try:
            # get a configerus LoadedConfig for the sonobuoy label
            loaded = self._environment.config().load(self._config_label)
        except ValidationError as err:
            raise ValueError("Invalid sonobuoy config received") from err

        # Get the run time arguments
        run_args: List[str] = []
        for (key, value) in loaded.get([self._config_base, "run"], default={}).items():
            run_args.append(f"--{key}={value}")

        return self.get_client_plugin().run(wait=wait, run_args=run_args)

    def status(self) -> SonobuoyStatus:
        """Retrieve Sonobuoy status return."""
        return self.get_client_plugin().status()

    def retrieve(self) -> SonobuoyResults:
        """Retrieve sonobuoy results."""
        logger.debug("retrieving sonobuoy results")
        return self.get_client_plugin().retrieve()

    def destroy(self, wait: bool = True):
        """Delete sonobuoy resources."""
        logger.debug("removing sonobuoy infrastructure")
        self.get_client_plugin().delete(wait=wait)

    def client_instance_id(self) -> str:
        """Return the instanceid for the child client plugin."""
        return f"{self._instance_id}-{METTA_SONOBUOY_CLIENT_PLUGIN_ID}"

    def get_client_plugin(self) -> SonobuoyClientPlugin:
        """Retrieve the client plugin if we can."""
        try:
            return self.fixtures.get_plugin(plugin_id=METTA_SONOBUOY_CLIENT_PLUGIN_ID)
        except KeyError as err:
            # Prepare was likely not run yet.
            raise RuntimeError(
                "Sonobuoy workload cannot find a client plugin. We "
                "likely have not created it yet. Did you run the "
                "prepare() method yet?"
            ) from err



"""

DISCOVER: Functionality for discovering METTA environments.

Discover is respondible for discovering project roots and initial project
configuration, along with related configuration, dependency injections and
fixtures.

"""
import os.path
import sys
import logging
from typing import List

from configerus.plugin import Type as ConfigerusType
from configerus.config import Config
from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH

logger = logging.getLogger("metta.discover")

METTA_CONFIG_LABEL = "metta"
""" this config label is used as a default for discovery from config """
METTA_CONFIG_SOURCES_KEY = "config.sources"
""" This config key is used to find additional config sources to add """
METTA_CONFIG_IMPORTS_KEY = "imports"
""" This config key is used to find modules to import """

METTA_ROOT_FILES = ["metta.yml", "metta.json"]
""" The presence of any of these files could be used to mark a conf source """

DEFAULT_SOURCE_PRIORITY = 40
""" Default priority for configerus path sources for root marker files """
DEFAULT_SOURCE_CONFIG_PRIORITY = 70
""" Default priority for configerus sources found in config """

DEFAULT_ENVIRONMENT_ROOT_CONFIG_IF_NO_ROOT_IS_FOUND = {
    "metta": {"project": {"name": "none"}},
    "environments": {"missing": {}},
}
""" If no root is found then this config is passed back as a config source """


def discover_project_root(config: Config, start_path: str, marker_files: List[str] = None):
    """Find a project root path.

    We start looking in the start_path for certain marker files, and if we
    don't find any then we check the parent, recursively. If we never find a
    marker file then we assume that the current path is the root.

    Whichever path we think is the root, we add a a configerus path source.
    That means that you can put whatever you want there for config.

    Commonly, one used just the metta.yml|json file to direct metta to include
    additional paths (such a ./config) from the metta.yml path.

    """
    if marker_files is None:
        marker_files = METTA_ROOT_FILES

    # standardize the path to rid of issues that could be cause by path tricks
    # such as  'path/..'
    start_path = os.path.realpath(start_path)

    # Try to any a path from start+path and up that contains a marker file as a
    # config source.
    check_path = start_path
    depth = 0
    while check_path:
        # if we are in a system /root path then stop scanning
        # we do this by checking if the path's parent is the
        # same as itself.
        if check_path == os.path.dirname(check_path):
            break

        for root_file in marker_files:
            root_file_path = os.path.join(check_path, root_file)

            if os.path.isfile(root_file_path):
                # If we found a marker file in a path, then we add that path as
                # a config source.
                if check_path not in sys.path:
                    sys.path.append(check_path)

                priority = DEFAULT_SOURCE_PRIORITY - depth
                if depth:
                    instance_id = f"project-{depth}"
                else:
                    instance_id = "project"
                config.add_source(
                    plugin_id=PLUGIN_ID_SOURCE_PATH,
                    instance_id=instance_id,
                    priority=priority,
                ).set_path(check_path)
                logger.info("Added project path as config: %s => %s", check_path, instance_id)

                depth += 1

                # Since we already added this path as a config source, we don't
                #  need to keep looking for more files.
                break

        # move up one directory and try again
        check_path = os.path.dirname(check_path)

    try:
        config.plugins.get_plugins(type=ConfigerusType.SOURCE)
    except KeyError:
        logger.warning("No project config found")

    return config



"""

On bootstrap, add some common config values that can be useful to a project,
such as some information about the current user and current path, platform and
a fixed single datetime for operation.

This is optional, and may conflict or cloud some ways of using metta, so it is
only meant to be used when the infomration is useful.

"""
import os
import logging
import getpass
from datetime import datetime
import platform

import appdirs

from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH as CONFIGERUS_SOURCE_PATH
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT as CONFIGERUS_SOURCE_DICT
from mirantis.testing.metta.environment import Environment

logger = logging.getLogger("metta.common")

DIR = os.path.abspath(os.getcwd())

METTA_COMMON_APP_NAME = "metta"
""" used for some path building for config sources """
METTA_COMMON_CONFIG_USER_INSTANCE_ID = "user"
""" config source instance id for user path """
METTA_COMMON_CONFIG_PROJECT_DYNAMIC_INSTANCE_ID = "project_dynamic"
""" config source instance id for dynamic dict of commonly used value """
METTA_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS = 35
""" Config source priority for added common config """
METTA_COMMON_PROJECT_CONFIG_SUBPATH = "config"
""" convention for configuration path as a subfolder of the cwd path """

MARKER_FILES = {
    "metta.py",
    "mettac.py",
    "conftest.py",
    "pytest.ini",
}
""" Files which mark the root of a project """


def add_common_config(environment: Environment):
    """Add some common configuration sources.

    The following could be added:

    1. a `${PWD}/config` path source named "project_config" if that path exists
    2. a user defaults `~/metta` path source if that path exists
    3. a dynamic dict of common values:
        i. `user:id` pulled from the os
        ii. `global:datetime` to provide a constant datetime that can be used
           for consistent labelling
        iii `path:project` path to ${PWD} in case you want to build file paths
           based on that path

        @note we will likely grow this with common sense values

    Parameters:
    -----------
    environment (Environment) : METTA environment to be modified

    """

    # a user config path (like ~/.config/metta) may contain config
    user_conf_path = appdirs.user_config_dir(METTA_COMMON_APP_NAME)
    if not environment.config().has_source(METTA_COMMON_CONFIG_USER_INSTANCE_ID) and os.path.isdir(
        user_conf_path
    ):
        environment.config().add_source(
            CONFIGERUS_SOURCE_PATH,
            METTA_COMMON_CONFIG_USER_INSTANCE_ID,
            METTA_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS,
        ).set_path(user_conf_path)

    # Add some dymanic values for config
    environment.config().add_source(
        CONFIGERUS_SOURCE_DICT,
        METTA_COMMON_CONFIG_PROJECT_DYNAMIC_INSTANCE_ID,
        priority=METTA_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS,
    ).set_data(
        {
            "user": {"id": getpass.getuser()},  # override user id with a host value
            "environment": environment.info(),
            "global": {
                "datetime": datetime.now(),  # use a single datetime across all checks
            },
            environment.config().paths_label(): {  # config label for paths, usually just "paths"
                # you can use "paths:project" in config to substitute this path
                "project": find_project_root_path(DIR)
            },
            "platform": {
                "machine": platform.machine(),
                "system": platform.system(),
                "processor": platform.processor(),
                "version": platform.version(),
                "uname": platform.uname(),
            },
        }
    )


def find_project_root_path(start_path: str) -> str:
    """Find a project root path.

    We start looking in the cwd for certain marker files, and if we don't find
    any then we check the parent, recursively.

    If we never find a marker file then we assume that the current path is the
    root.

    """
    # Try to add a path from cwd and up that contains a mettac.py file
    check_path = start_path
    while check_path:
        if check_path == "/":
            break

        for marker_file in MARKER_FILES:
            marker_path = os.path.join(check_path, marker_file)
            if os.path.isfile(marker_path):
                return check_path

        check_path = os.path.dirname(check_path)

    return check_path



"""

Environment definitions and functionality.

An environment in Metta is a container for the metta components. It is created
early and used to access all of the metta functionality.  For testing systems,
you can think of an environment as a collection of interconnected resources to
be managed together. If components need to be started or stopped together then
they are probably a part of the same environment.
When using Metta you can use more than one Environment object, but often you
don't need to.

The two primary components of an environment are:
1. A configerus config object to be shared by functionality that identifies as
  being in the environment.  This includes Metta components, but is intended to
  be used by external components as well.
2. A Fixtures set of Metta Fixture/Plugin components created in the environment.

In this file are defined two Environment plugins:

1. Default : a bare bones environment object, which you can use in any project
             to manually create fixtures/plugins and access them.
2. Builder : an environment with a number builder functions added to make it
             easier to create plugins en masse, using configerus config or from
             dicts of data.
             This environment also uses config for importing more python modules
             and for adding configerus config sources at runtime.

Almost all of the Mirantis test suites use the builder or a derivative, and the
default environment is there primarily as a stripped down simple version.

"""
from typing import Dict, List, Any
from logging import getLogger

from configerus.config import Config
from configerus.loaded import LOADED_KEY_ROOT

from .plugin import (
    Factory,
)
from .fixture import (
    Fixture,
    Fixtures,
    METTA_FIXTURES_CONFIG_FIXTURES_LABEL,
)
from .building import FixtureBuildingFromConfigMixin, FixtureBuildingFromDictMixin
from .config import add_config_sources_from_config, METTA_CONFIG_CONFIG_SOURCE_KEY
from .importing import add_imports_from_config, METTA_IMPORT_CONFIG_LABEL
from .setuptools import setuptools_entrypoint, METTA_CONFIG_SETUPTOOLS_BOOTSTRAPS_KEY

logger = getLogger("metta.environment")

METTA_ENTRYPOINT_BOOTSTRAP_ENVIRONMENT = "metta.bootstrap.environment"
"""Setuptools entrypoint for boostrapping environments."""

METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENT = "environment"
""" metta plugin interface identifier for environment plugins """


METTA_FIXTURES_CONFIG_ENVIRONMENTS_KEY = "environments"
"""Configerus .get() key for finding environments."""


METTA_CORE_ENVIRONMENT_PLUGIN_ID = "metta_core_environment"
""" Metta plugin id for the base environment plugin.

In the Factory below, this ID is used to register the factory function which
builds an instance of the "Environment" class, defined a few lines below.

That Environment plugin is meant to be used as a simple Environment
implementation with enough functionality to be fully usable. See also the
Builder plugin which has more code to generate plugins automatically.

"""


@Factory(
    plugin_id=METTA_CORE_ENVIRONMENT_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENT],
)
def core_environment_factory(config: Config, instance_id: str) -> "Environment":
    """Build a FixtureBuilderEnvironment environment."""
    return Environment(config=config, instance_id=instance_id)


class Environment:
    """Environment with helpers for building fixtures.

    The building Environment handler is a straight forward bare Environment
    object with additional helper methods to build fixtures in the Environment.

    This separates out the two base needs for an environment:
    1. the simple raw interactive functionality for the Environment,
    2. the magic config and Dict interpration for building Fixtures

    The combination makes this class quite bare, but allows other Environment
    objects to add on more functionality without creating import circuits.
    """

    def __init__(self, config: Config, instance_id: str):
        """Get an environment name and a config object.

        Parameters:
        -----------
        instance_id (str) : Arbitrary string label for the environment, not used for
            any functionality, and so not limited in uniqueness or syntax.
        config (Config) : Configerus Config object used internally and by any
            functionality that wants to consider itself inside the environment.

        """
        self._instance_id: str = instance_id
        """Arbitrary label for the environment, not used functionally."""
        self._config: Config = config
        """Config object usable by any component that is in the env."""
        self._fixtures: Fixtures = Fixtures()
        """An environment is a container of Fixture objects."""

    # pylint: disable=unused-argument
    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Return dict plugin info."""
        return {
            "name": self.instance_id(),
        }

    def set_state(self, state: str):
        """Indicate to an Environment what state is expected.

        This is an injection option which needs to be at this base object, but
        is meant to allow different Environment plugin classes to respond in
        any manner.
        The method has a very simple interface, because it needs to be generic.

        This was develloped with the contrib state based environments in mind,
        and is actually a leftover from the previous state design, but I don't
        have an abstract way to pull it out without big overhauls and even more
        abstractions and overrides.  Right now this facilitates the cli control
        for state.

        """
        logger.warning(
            "Environment %s doesn't accomodate state changes: %s", self._instance_id, state
        )

    def instance_id(self) -> str:
        """Get the string label for the environment.

        Returns:
        --------
        String label for the environment
        """
        return self._instance_id

    def config(self) -> Config:
        """Configerus Config access for the environment.

        Returns:
        --------
        A Configerus Config object which is meant to be used by anything inside
        the environment.
        """
        return self._config

    def fixtures(self) -> Fixtures:
        """Fixtures collection accessor for the environment.

        Returns:
        --------
        A Fixtures object which is a set of all Fixture/Plugin compoents in the
        environment.
        """
        return self._fixtures

    # This is what it takes to build a plugin
    # pylint: disable=too-many-arguments
    def new_fixture(
        self,
        plugin_id: str,
        instance_id: str,
        priority: int,
        arguments: Dict[str, Any] = None,
        labels: Dict[str, Any] = None,
        replace_existing=False,
    ) -> Fixture:
        """Create a new plugin from parameters.

        Parameters:
        -----------
        plugin_id (str) : METTA plugin id to tell us what plugin factory to use;

            @see .plugin.Factory for more details on how plugins are loaded.

        instance_id (str) : string instance id that will be passed to the new
            plugin object;

        priority (int) : Integer priority 1-100 for comparative prioritization
            between other plugins;

        arguments (Dict[str, Any]) : Keyword Arguments which should be passed to
            the plugin constructor after environment and instance_id;

        labels (Dict[str, str]) : Keyword/value labels which are to be associated
            with the plugin/fixture;

        replace_existing (bool) : Replace any existing matching fixture.

        Return:
        -------
        A Fixture object with the new plugin added

        The Fixtures has already been added to the environment, but is returned
        so that the consumer can act on it separately without haveing to
        search for it.

        Raises:
        -------
        NotImplementedError if you asked for an unregistered plugin_id

        """
        if arguments is None:
            arguments = {}

        # Catch some early arg validation errors which would otherwise have
        # caused some hard to diagnose issues.
        if not (
            isinstance(plugin_id, str)
            and isinstance(instance_id, str)
            and isinstance(priority, int)
        ):
            raise ValueError(
                f"Bad arguments passed for creating a fixture: "
                f":{plugin_id}:{instance_id} ({priority})"
            )

        if labels is None:
            labels = {}
        labels["environment"] = self.instance_id()

        # Build the plugin instance by passing collected arguments to the
        # registered plugin factory. This will call whatever function was
        # decorated for the plugin_id.
        args: List[Any] = [self, instance_id]
        kwargs: Dict[str, Any] = arguments
        plugin_instance = Factory.create(plugin_id, instance_id, *args, **kwargs)

        # Build a fixture from the plugin_instance and add it to the fixtures
        # set for the environment, then return the fixture.
        fixture = self.fixtures().add(
            fixture=Fixture.from_instance(plugin_instance, priority=priority, labels=labels),
            replace_existing=replace_existing,
        )
        return fixture


METTA_BUILDER_ENVIRONMENT_PLUGIN_ID = "metta_builder_environment"
""" Metta plugin id for the classic environment plugin.

In the Factory below, this ID is used to register the factory function which
builds an instance of the "FixtureBuilderEnvironment" class, defined a few
lines below.

The Builder environment includes code for more automated inclusion, configuration
and creation of environment plugins, either from config or from dicts.
If you like config based environments, this is likely what you want.
"""


@Factory(
    plugin_id=METTA_BUILDER_ENVIRONMENT_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENT],
)
def builder_environment_factory(
    config: Config, instance_id: str, label: str = "", base: str = LOADED_KEY_ROOT
) -> "FixtureBuilderEnvironment":
    """Build a FixtureBuilderEnvironment environment."""
    return FixtureBuilderEnvironment(
        config=config, instance_id=instance_id, label=label, base=base
    )


class FixtureBuilderEnvironment(
    Environment, FixtureBuildingFromConfigMixin, FixtureBuildingFromDictMixin
):
    """Environment with helpers for building fixtures.

    The building Environment handler is a straight forward bare Environment
    object with additional helper methods to build fixtures in the Environment.

    This separates out the the magic config and Dict interpration for building
    Fixtures from the base Environment object.  Both can be used, but this one
    attempts to make plugin generation easier.

    The combination makes this class quite bare, but allows other Environment
    objects to add on more functionality without creating import circuits.
    """

    def __init__(
        self, config: Config, instance_id: str, label: str = "", base: str = LOADED_KEY_ROOT
    ):
        """Get an environment name and a config object.

        Parameters:
        -----------
        name (str) : Arbitrary string label for the environment, not used for
            any functionality, and so not limited in uniqueness or syntax.
        config (Config) : Configerus Config object used internally and by any
            functionality that wants to consider itself inside the environment.
        """
        Environment.__init__(self, config=config, instance_id=instance_id)
        FixtureBuildingFromConfigMixin.__init__(
            self, config=config, builder_callback=self.new_fixture
        )
        FixtureBuildingFromDictMixin.__init__(self, builder_callback=self.new_fixture)

        self._environment_boostraps: List[str] = []
        """Track what bootstaps we have applied for introspection."""

        # If we received any config directiosn, then we self-bootstrap from the config
        if label:
            env_config = config.load(label)

            # 1. add any config sources specified in config
            # (Here config tells us to load more config. It is weird but ok.)
            add_config_sources_from_config(
                config=config, label=label, base=[base, METTA_CONFIG_CONFIG_SOURCE_KEY]
            )

            # 2. import any python code requested, to make sure that all funxtionality
            #    is in scope that is needed.
            #    Here we say look in the "metta" config for a "imports" section.
            add_imports_from_config(
                config=config, label=label, base=[base, METTA_IMPORT_CONFIG_LABEL]
            )

            # 3. run an setuptools bootstrap entrypoints
            self._environment_boostraps = env_config.get(
                [base, METTA_CONFIG_SETUPTOOLS_BOOTSTRAPS_KEY], default=[]
            )
            setuptools_entrypoint(
                entrypoint=METTA_ENTRYPOINT_BOOTSTRAP_ENVIRONMENT,
                entries=self._environment_boostraps,
                args=[self],
                kwargs={},
            )

            # 4. build fixtures from config
            try:
                labels: Dict[str, str] = {
                    "container": "environment",
                    "container_id": self.instance_id(),
                    "environment": self.instance_id(),
                }
                self.add_fixtures_from_config(
                    label=label,
                    base=[base, METTA_FIXTURES_CONFIG_FIXTURES_LABEL],
                    labels=labels,
                )
            except KeyError as err:
                logger.warning(
                    "%s Environment plugin encountered an issue creating fixtures: %s",
                    self.__class__,
                    err,
                )

    # pylint: disable=unused-argument
    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Return dict plugin info."""
        return {
            "name": self.instance_id(),
            "bootstraps": self._environment_boostraps,
        }



"""

Metta client plugin for sonobuoy.

Metta client plugin which provides an injectable metta plugin for interacting
with the SonobuoyClient code with an applied set of configuration.
These are typically provided by the other plugins, but could be created
directly if you can pass the arguments in.

"""
from typing import List, Dict
import subprocess

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta_health.healthcheck import Health

from mirantis.testing.metta_kubernetes.kubeapi_client import KubernetesApiClientPlugin

from .sonobuoy import SonobuoyClient, SONOBUOY_DEFAULT_RESULTS_PATH
from .plugin import Plugin
from .results import (
    Status,
    SonobuoyStatus,
    SonobuoyResults,
)

METTA_SONOBUOY_CLIENT_PLUGIN_ID = "metta_sonobuoy_client"
""" workload plugin_id for the sonobuoy plugin """


class SonobuoyClientPlugin:
    """Metta client plugin for sonobuoy."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        kubeclient: KubernetesApiClientPlugin,
        plugins: List[Plugin] = None,
        results_path: str = SONOBUOY_DEFAULT_RESULTS_PATH,
    ):
        """Gather enough arguments to configure the SonobuoyClient object."""
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self._sonobuoy: SonobuoyClient = SonobuoyClient(
            kubeclient=kubeclient,
            plugins=plugins,
            results_path=results_path,
        )

    # the deep argument is a standard for the info hook
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Return dict data about this plugin for introspection."""
        return {
            "client": self._sonobuoy.info(deep=deep) if self._sonobuoy is not None else "MISSING"
        }

    def run(self, wait: bool = True, run_args: List[str] = None):
        """Run sonobuoy."""
        return self._sonobuoy.run(wait=wait, run_args=run_args)

    def status(self) -> SonobuoyStatus:
        """Retrieve Sonobuoy status return."""
        return self._sonobuoy.status()

    def logs(self, follow: bool = True):
        """Output sonobuoy logs."""
        return self._sonobuoy.logs(follow=follow)

    def results(self) -> str:
        """Retrieve sonobuoy results."""
        return self._sonobuoy.results()

    def retrieve(self) -> SonobuoyResults:
        """Retrieve sonobuoy results."""
        return self._sonobuoy.retrieve()

    def delete(self, wait: bool = False):
        """Delete sonobuoy resources."""
        return self._sonobuoy.delete(wait=wait)

    def version(self) -> Dict[str, str]:
        """Retrieve sonobuoy version."""
        return self._sonobuoy.version()

    def health(self) -> Health:
        """Perform a health check on the workload."""
        health = Health(source=self._instance_id)

        try:
            status = self.status()

            if status.status in [Status.POSTPROCESS]:
                health.info("Sonobuoy: run has finished, but result is not yet avaialble.")
            elif status.status in [Status.COMPLETE, Status.PASSED]:
                health.info("Sonobuoy: completed.")
            elif status.status in [Status.FAILED]:
                health.error("Sonobuoy: run has produced a failure.")
            else:  # if status.status() in [Status.PENDING, Status.RUNNING]:
                health.info("Sonobuoy: Running")

        except (subprocess.CalledProcessError, AttributeError) as err:
            health.unknown(f"No status found. Sonobuoy is likely not running: {err}")

        return health

    def create_k8s_crb(self):
        """Create the cluster role binding that sonobuoy needs."""
        return self._sonobuoy.create_k8s_crb()

    def delete_k8s_crb(self):
        """Remove the cluster role binding that we created."""
        return self._sonobuoy.delete_k8s_crb()



"""

METTA Cli package.

"""

import logging

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment

from .base import METTA_PLUGIN_INTERFACE_ROLE_CLI
from .config import ConfigCliPlugin
from .environment import EnvironmentCliPlugin
from .fixtures import FixturesCliPlugin
from .provisioner import ProvisionerCliPlugin
from .output import OutputCliPlugin

logger = logging.getLogger("metta.cli")


METTA_PLUGIN_ID_CLI_CONFIG = "config_cli"
""" cli plugin_id for the config plugin """


@Factory(plugin_id=METTA_PLUGIN_ID_CLI_CONFIG, interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI])
def metta_plugin_factory_cli_config(environment: Environment, instance_id: str = ""):
    """Create a config cli plugin."""
    return ConfigCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_ENVIRONMENT = "environment_cli"
""" cli plugin_id for the environment plugin """


@Factory(
    plugin_id=METTA_PLUGIN_ID_CLI_ENVIRONMENT,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI],
)
def metta_plugin_factory_cli_environment(environment: Environment, instance_id: str = ""):
    """Create an environment cli plugin."""
    return EnvironmentCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_FIXTURES = "fixtures_cli"
""" cli plugin_id for the fixtures plugin """


@Factory(plugin_id=METTA_PLUGIN_ID_CLI_FIXTURES, interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI])
def metta_plugin_factory_cli_fixtures(environment: Environment, instance_id: str = ""):
    """Create a fixtures cli plugin."""
    return FixturesCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_PROVISIONER = "provisioner_cli"
""" cli plugin_id for the provisioner plugin """


@Factory(
    plugin_id=METTA_PLUGIN_ID_CLI_PROVISIONER,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI],
)
def metta_plugin_factory_provisioner_config(environment: Environment, instance_id: str = ""):
    """Create a provisioner cli plugin."""
    return ProvisionerCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_OUTPUT = "output_cli"
""" cli plugin_id for the output plugin """


@Factory(
    plugin_id=METTA_PLUGIN_ID_CLI_OUTPUT,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI],
)
def metta_plugin_factory_output_config(environment: Environment, instance_id: str = ""):
    """Create a output cli plugin."""
    return OutputCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint BootStrapping ------


# pylint: disable=unused-argument
def bootstrap_environment(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added to.

    """



"""

Metta CLI plugin for launchpad

Provided CLI commands that allow interacting with launchpad plugins.
CLI commands exist for both provisioner and direct client plugins, with
some overlap existing.

"""

import logging
from typing import List, Dict

import yaml

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixture
from mirantis.testing.metta_cli.base import CliBase, cli_output

from .provisioner import (
    LaunchpadProvisionerPlugin,
    METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID,
)
from .client import (
    LaunchpadClientPlugin,
    METTA_LAUNCHPAD_CLIENT_PLUGIN_ID,
)

logger = logging.getLogger("metta.cli.launchpad")

METTA_LAUNCHPAD_CLI_PLUGIN_ID = "metta_launchpad_cli"
""" metta plugin_id for the launchpad cli plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class LaunchpadCliPlugin(CliBase):
    """Fire command/group generator for launchpad plugin commands."""

    def fire(self):
        """Return a dict of commands.

        Don't return any commands if there are no plugind available

        """
        commands: Dict[str, object] = {}

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_LAUNCHPAD_CLIENT_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            commands["launchpad"] = LaunchpadClientGroup(self._environment)

        return commands


class LaunchpadClientGroup:
    """Base Fire command group for launchpad client cli commands."""

    def __init__(self, environment: Environment):
        """Create launchpad command list object."""
        self._environment: Environment = environment

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            self.provisioner = LaunchpadProvisionerGroup(self._environment)

    def _select_client(self, instance_id: str = "") -> Fixture:
        """Pick a matching client."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=METTA_LAUNCHPAD_CLIENT_PLUGIN_ID,
                instance_id=instance_id,
            )

        # Get the highest priority client
        return self._environment.fixtures().get(plugin_id=METTA_LAUNCHPAD_CLIENT_PLUGIN_ID)

    def info(self, client: str = "", deep: bool = False):
        """Get info about a client plugin."""
        fixture = self._select_client(instance_id=client)
        return cli_output(fixture.info(deep=deep))

    def hosts(self, client: str = "", deep: bool = False):
        """List the hosts in the cluster."""
        client_plugin: LaunchpadClientPlugin = self._select_client(instance_id=client).plugin
        return cli_output(client_plugin.hosts(deep=deep))

    def exec(self, cmd: str, client: str = "", host: int = 0):
        """Exec a command."""
        client_plugin: LaunchpadClientPlugin = self._select_client(instance_id=client).plugin
        return client_plugin.exec(host_index=host, cmds=cmd.split(" "))

    def client_config(self, client: str = ""):
        """Get the rendered config from the client."""
        client_plugin: LaunchpadClientPlugin = self._select_client(instance_id=client).plugin
        return cli_output(client_plugin.describe_config())

    def version(self, client: str = ""):
        """Output a launchpad cli report."""
        client_plugin: LaunchpadClientPlugin = self._select_client(instance_id=client).plugin
        client_plugin.version()

    def describe(self, report: str, client: str = ""):
        """Output a launchpad cli report."""
        client_plugin: LaunchpadClientPlugin = self._select_client(instance_id=client).plugin
        client_plugin.describe(report)

    # pylint: disable=redefined-builtin
    def fixtures(
        self,
        client: str = "",
        plugin_id: str = "",
        interfaces: List[str] = None,
        instance_id: str = "",
    ):
        """List all outputs."""
        client_plugin: LaunchpadClientPlugin = self._select_client(instance_id=client).plugin

        fixture_list = [
            fixture.info()
            for fixture in client_plugin.fixtures().filter(
                interfaces=interfaces, plugin_id=plugin_id, instance_id=instance_id
            )
        ]

        return cli_output(fixture_list)

    def config_file(self, client: str = ""):
        """Dump the config file."""
        client_plugin: LaunchpadClientPlugin = self._select_client(instance_id=client).plugin

        try:
            with open(client_plugin.config_file, encoding="utf8") as config_file:
                config_contents = yaml.load(config_file)
        except FileNotFoundError as err:
            raise ValueError("No config file was found") from err

        return cli_output(config_contents)

    def client_bundle(self, client: str = "", user: str = "admin", reload: bool = False):
        """Tell Launchpad to download the client bundle."""
        client_plugin = self._select_client(instance_id=client)
        plugin = client_plugin.plugin

        logger.info("Downloading client bungle for user: %s", user)
        try:
            # access private method for manual interaction
            # pylint: disable=protected-access
            bundle = plugin._mke_client_bundle(user=user, reload=reload)
        except Exception as err:
            raise Exception("Launchpad command failed.") from err

        return cli_output(bundle)

    def apply(self, debug: bool = False, client: str = ""):
        """Run client apply."""
        client_plugin: LaunchpadClientPlugin = self._select_client(instance_id=client).plugin
        client_plugin.apply(debug=debug)

    def reset(self, client: str = ""):
        """Run client destroy."""
        client_plugin: LaunchpadClientPlugin = self._select_client(instance_id=client).plugin
        client_plugin.reset()


class LaunchpadProvisionerGroup:
    """Launchpad cli commands."""

    def __init__(self, environment: Environment):
        """Create launchpad command list object."""
        self._environment: Environment = environment

    def _select_provisioner(self, instance_id: str = "") -> Fixture:
        """Pick a matching provisioner."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID,
                instance_id=instance_id,
            )

        # Get the highest priority provisioner
        return self._environment.fixtures().get(plugin_id=METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID)

    def info(self, provisioner: str = "", deep: bool = False):
        """Get info about a provisioner plugin."""
        fixture = self._select_provisioner(instance_id=provisioner)
        return cli_output(fixture.info(deep=deep))

    def prepare(self, provisioner: str = ""):
        """Run provisioner prepare."""
        provisioner_plugin: LaunchpadProvisionerPlugin = self._select_provisioner(
            instance_id=provisioner
        ).plugin
        provisioner_plugin.prepare()

    def apply(self, debug: bool = False, provisioner: str = ""):
        """Run provisioner apply."""
        provisioner_plugin: LaunchpadProvisionerPlugin = self._select_provisioner(
            instance_id=provisioner
        ).plugin
        provisioner_plugin.apply(debug=debug)

    def destroy(self, provisioner: str = ""):
        """Run provisioner destroy."""
        provisioner_plugin: LaunchpadProvisionerPlugin = self._select_provisioner(
            instance_id=provisioner
        ).plugin
        provisioner_plugin.destroy()

    def write_config(self, provisioner: str = ""):
        """Write config to file."""
        provisioner_plugin: LaunchpadProvisionerPlugin = self._select_provisioner(
            instance_id=provisioner
        ).plugin
        # access private method for introspection
        # pylint: disable=protected-access
        provisioner_plugin._write_launchpad_yml()



"""

Metta client plugin for launchpad.

Primarily a metta plugin interface for the ./launchpad subprocess client, but in a form
that acts as a metta plugin.  This means that it can form a part of an metta environment.

Most other plugins that use launchpad will themselves proxy to this plugin.

"""
import logging
from typing import Dict, Any, List
import subprocess

from configerus.loaded import Loaded

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures

from mirantis.testing.metta_mirantis.mke_client import (
    MKEAPIClientPlugin,
    METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID,
)
from mirantis.testing.metta_mirantis.msr_client import METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID

from .launchpad import (
    LaunchpadClient,
    METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT,
    METTA_LAUNCHPADCLIENT_WORKING_DIR_DEFAULT,
)

logger = logging.getLogger("metta_launchpad:provisioner")

METTA_LAUNCHPAD_CLIENT_PLUGIN_ID = "metta_launchpad_client"
""" Metta plugin_id for the launchpad provisioner plugin """

METTA_LAUNCHPAD_CONFIG_HOSTS_KEY = "spec.hosts"
""" config key for the list of cluster hosts. """
METTA_LAUNCHPAD_CONFIG_HOST_ROLE_KEY = "role"
""" config key for the role of a host """

METTA_LAUNCHPAD_CLI_CONFIG_DOCKER_VERSION_DEFAULT = "1.40"
""" Default value for the docker client version number."""


# pylint: disable=too-many-arguments
class LaunchpadClientPlugin:
    """Metta client plugin for launchpad."""

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        config_file: str = METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT,
        working_dir: str = METTA_LAUNCHPADCLIENT_WORKING_DIR_DEFAULT,
        cli_options: Dict[str, bool] = None,
        systems: Dict[str, Dict[str, str]] = None,
    ):
        """Collect enough data to create a LaunchpadClient object.

        Parameters:
        -----------
        config_file (str) : Path to the launchpad yml config file.

        working_dir (str) : Path CWD to use for subprocess with launchpad.  This may be needed
            if PEMs in the yaml are relative paths.

        cli_options (Dict[str, Any]) : Additional -- flags which should be passed to all
            launchpad commands.

        systems (Dict[str, Any]) : Dictionary to provide client generation for infrastructure
            created by Launchpad.  The two primary examples are MKE and MSR.

            Each key of the Dict is the identifier of the system, and its values are arguments
            to the constructor of the system, except the host list.

            An example for MKE:
                "mke": {
                    "accesspoint": "192.168.172.11",
                    "username": "admin"",
                    "password": "orca",
                }

                A host list will be added to the arguments.

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self.config_file: str = config_file
        """Path to the launchpad yml file."""

        self.systems: Dict[str, Dict[str, str]] = systems if systems is not None else {}
        """Access endpoint & U/P for systems created by launchpad, such as the MKE client."""

        self._fixtures: Fixtures = Fixtures()
        """This plugin makes fixtures, and keeps track of them here."""

        logger.debug("Creating Launchpad client handler")
        self.launchpad: LaunchpadClient = LaunchpadClient(
            config_file=config_file,
            working_dir=working_dir,
            cli_options=cli_options,
        )

        # If we can, it makes sense to build the MKE and MSR client fixtures now.
        # This will only be possible in cases where we have an installed cluster.
        # We try that here, even though it is verbose and ugly, so that we have
        # the clients available for introspection, for all consumers.
        # We probably shouldn't, but it allows some flexibility.
        # attempt to be declarative and make the client plugin in case the
        # terraform chart has already been run.
        # self.make_fixtures()
        try:
            self.make_fixtures()

        # dont' block the construction on an exception
        # pylint: disable=broad-except
        except Exception:
            pass

    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Get info about a provisioner plugin.

        Returns:
        --------
        Dict of introspective information about this plugin_info

        """
        return {
            "config": {
                "launchpad_yml_path": self.config_file,
            },
            "systems": self.systems,
            "client": self.launchpad.info(deep=deep),
        }

    def fixtures(self) -> Fixtures:
        """Return children fixtures of the client."""
        return self._fixtures

    def version(self):
        """Get the launchpad version."""
        return self.launchpad.version()

    def hosts(self, deep: bool = False):
        """List the hosts in the cluster."""
        config = self.describe_config()

        if deep:
            host_list = config["spec"]["hosts"]
        else:
            host_list = []
            for host in config["spec"]["hosts"]:
                list_host = {"role": host["role"]}
                if "ssh" in host:
                    list_host.update({"is_windows": False, "address": host["ssh"]["address"]})
                if "winrm" in host:
                    list_host.update({"is_windows": True, "address": host["winrm"]["address"]})

                host_list.append(list_host)

        return host_list

    def apply(self, debug: bool = False):
        """Bring a cluster up.

        We assume that the cluster is running and the we can pull the required
        yaml from an output fixture in the environment.

        This plugin needs an output fixture, probably of dict type.  It will
        Pull that structure for the launchpad yaml config file and dump it into
        its config path.
        The provisioner can find an output directly from the environment, or
        from a specific fixture source.  If you want the output to come from
        only a specific backend fixture then make sure that a "backend" config
        exists, otherwise just use an "output" config.

        Raises:
        -------
        ValueError if the object has been configured (prepare) with config that
            doesn't work, or if the backend doesn't give valid yml

        Exception if launchpad fails.

        """
        logger.info("Using launchpad to install products onto backend cluster")
        self.launchpad.apply(debug=debug)
        self.make_fixtures()

        try:
            mke_plugin = self._get_mke_client_plugin()
            # We got an MKE client, so let's activate it.
            mke_plugin.api_get_bundle(force=True)
            mke_plugin.make_fixtures()
        except KeyError as err:
            raise RuntimeError("Launchpad MKE client failed to download client bundle.") from err

        # as we have likely changed our cluster, tell everybody to download new
        # client bundles.
        try:
            # this plugin client maintains a client bundle
            self.launchpad.get_bundle(user="admin")
        except KeyError as err:
            raise RuntimeError("Launchpad client failed to download client bundle.") from err

    def reset(self):
        """Ask the client to remove installed resources."""
        # tell the MKE client to remove its bundles
        try:
            mke = self._fixtures.get_plugin(
                plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID,
            )
            mke.rm_bundle()

        except KeyError as err:
            logger.warning("Launchpad's MKE plugin failed do remove a client bundle: %s", err)

        # now tell the launchpad client to reset
        try:
            self.launchpad.reset()
            self.launchpad.rm_client_bundles()

        except subprocess.CalledProcessError as err:
            logger.warning("Launchpad failed to destroy installed resources: %s", err)

    def bundle(self, user: str = "admin", reload: bool = False) -> Dict[str, Any]:
        """Retrieve the user bundle for the marked user and return an info dict about it."""
        self.launchpad.get_bundle(user=user, reload=reload)
        return self.launchpad.bundle(user=user)

    def register(self, name: str, email: str, company: str):
        """Uninstall using the launchpad client."""
        return self.launchpad.register(name=name, email=email, company=company)

    def describe(self, report: str):
        """Output one of the launchpad reports."""
        return self.launchpad.describe(report=report)

    def describe_config(self) -> Dict[str, Any]:
        """Return the launchpad config report as interpreted by launchpad."""
        return self.launchpad.describe_config()

    def exec(self, host_index: int, cmds: List[str]):
        """Execute a command on a host index."""
        return self.launchpad.exec(host_index=host_index, cmds=cmds)

    def make_fixtures(self) -> Fixtures:
        """Build fixtures for all of the clients.

        Returns:
        --------
        Fixtures collection of fixtures that have been created

        """
        # get fresh values for the launchpad config (in case it has changed) and
        # treat this as a configerus Loaded object which allows us to use the
        # searching and validation syntax.
        client_config = self.describe_config()
        launchpad_config = Loaded(data=client_config, parent=None, instance_id="launchpad_client")

        # Retrieve a list of hosts, and use that to decide what clients to
        # make.  If we find a host for a client, then we retrieve needed
        # config and use it to generate the related client.
        hosts = launchpad_config.get(METTA_LAUNCHPAD_CONFIG_HOSTS_KEY, default=[], format=False)

        # MKE Client
        #
        if METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID in self.systems:
            mke_hosts: List[Any] = list(
                host for host in hosts if host[METTA_LAUNCHPAD_CONFIG_HOST_ROLE_KEY] in ["manager"]
            )
            if len(mke_hosts) > 0:
                instance_id = f"{self._instance_id}-{METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID}"
                mke_arguments: Dict[str, Any] = self.systems[METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID]
                mke_arguments["hosts"] = mke_hosts

                if "accesspoint" in mke_arguments and mke_arguments["accesspoint"]:
                    mke_arguments["accesspoint"] = clean_accesspoint(mke_arguments["accesspoint"])

                logger.debug("Launchpad client is creating an MKE client plugin: %s", instance_id)
                fixture = self._environment.new_fixture(
                    plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID,
                    instance_id=instance_id,
                    priority=70,
                    arguments=mke_arguments,
                    labels={
                        "parent_plugin_id": METTA_LAUNCHPAD_CLIENT_PLUGIN_ID,
                        "parent_instance_id": self._instance_id,
                    },
                    replace_existing=True,
                )
                self._fixtures.add(fixture, replace_existing=True)

        # MSR Client
        #
        if METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID in self.systems:
            msr_hosts: List[Any] = list(
                host for host in hosts if host[METTA_LAUNCHPAD_CONFIG_HOST_ROLE_KEY] in ["msr"]
            )
            if len(msr_hosts) > 0:
                instance_id = f"{self._instance_id}-{METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID}"
                msr_arguments: Dict[str, Any] = self.systems[METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID]
                msr_arguments["hosts"] = msr_hosts

                if "accesspoint" in msr_arguments and msr_arguments["accesspoint"]:
                    msr_arguments["accesspoint"] = clean_accesspoint(msr_arguments["accesspoint"])

                logger.debug("Launchpad client is creating an MSR client plugin: %s", instance_id)
                fixture = self._environment.new_fixture(
                    plugin_id=METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID,
                    instance_id=instance_id,
                    priority=70,
                    arguments=msr_arguments,
                    labels={
                        "parent_plugin_id": METTA_LAUNCHPAD_CLIENT_PLUGIN_ID,
                        "parent_instance_id": self._instance_id,
                    },
                    replace_existing=True,
                )
                self._fixtures.add(fixture, replace_existing=True)

    def _get_mke_client_plugin(self) -> MKEAPIClientPlugin:
        """Retrieve the MKE client plugin if we can."""
        try:
            return self._fixtures.get_plugin(plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID)
        except KeyError as err:
            raise RuntimeError(
                "Launchpad client cannot find its MKE client plugin, and "
                "cannot process any client actions.  Was a client created?"
            ) from err


def clean_accesspoint(accesspoint: str) -> str:
    """Remove any https:// and end / from an accesspoint."""
    accesspoint = accesspoint.replace("https://", "")
    return accesspoint



"""

Metta client plugins for the ansible cli handlers.

"""

import logging
from typing import List, Dict

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta_health.healthcheck import Health

from .ansiblecli import AnsibleClient, AnsiblePlaybookClient

logger = logging.getLogger("metta.contrib.client:ansible")

METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID = "metta_ansible_clicore_ansible_client"
""" Ansible client plugin id """

METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKCLIENT_PLUGIN_ID = "metta_ansible_clicore_ansibleplaybook_client"
""" Ansible-Playbook client plugin id """


# pylint: disable=too-many-instance-attributes
class AnsibleClientPlugin:
    """Ansible provisioner plugin.

    Client plugin that allows control of and interaction with a ansible
    cluster using ansible cli.

    ## Requirements

    1. this plugin uses subprocess to call a ansible binary, so you have to install
       ansible in the environment

    ## Usage

    """

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        inventory_path: str,
        ansiblecfg_path: str = "",
    ):
        """Initialize Ansible Playbook client.

        Parameters:
        -----------
        environment (Environment) : All metta plugins receive the environment
            object in which they were created.
        instance_id (str) : all metta plugins receive their own string identity.

        inventory_path (str) : string path to an inventory file.
        ansiblecfg_path (str) : string path to an optional ansiblecfg file.

        """
        self._environment: Environment = environment
        """Environemnt in which this plugin exists."""
        self._instance_id: str = instance_id
        """Unique id for this plugin instance."""

        self._ansible: AnsibleClient = AnsibleClient(
            ansiblecfg_path=ansiblecfg_path,
            inventory_path=inventory_path,
        )
        """Backend ansible cli handler."""

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Get info about a provisioner plugin."""
        return {
            "client": self._ansible.info(deep=deep) if self._ansible else "MISSING",
        }

    def health(self) -> Health:
        """Return health status of the cluster."""
        ansible_health = Health(source=self._instance_id)

        if self._ansible:
            for test_health_function in [
                self._health_all_ping,
            ]:
                test_health = test_health_function()
                ansible_health.merge(test_health)

        return ansible_health

    def run(
        self,
        args: List[str],
        envs: Dict[str, str] = None,
        return_output: bool = False,
    ):
        """Run a set of string ansible arguments."""
        return self._ansible.run(args=args, envs=envs, return_output=return_output)

    def debug(self, hosts: str = "all"):
        """Run a set of string ansible arguments."""
        return self._ansible.debug(hosts=hosts)

    def setup(self, hosts: str = "all"):
        """Run a set of string ansible arguments."""
        return self._ansible.setup(hosts=hosts)

    def ping(self, hosts: str = "all"):
        """Run a set of string ansible arguments."""
        return self._ansible.ping(hosts=hosts)

    def _health_all_ping(self) -> Health:
        """Health check that tries to ping all of the hosts."""
        ping_health = Health(source=self._instance_id)

        ping = self.ping()

        try:
            ping_task_result_hosts = ping["plays"][0]["tasks"][0]["hosts"]
            stats_hosts = ping["stats"]
        except KeyError:
            ping_health.error("ansible ping gave unexpected results.")
        else:
            for host, host_stats in stats_hosts.items():
                if host_stats["ok"]:
                    ping_health.healthy(f"Ansible: {host} ping response ok.")
                elif host_stats["unreachable"]:
                    ping_health.warning(f"Ansible: {host} unreachable during ping.")
                elif host_stats["ignored"]:
                    ping_health.warning(f"Ansible: {host} ping ignored.")
                elif host_stats["failures"]:
                    ping_health.error(f"Ansible: {host} ping failed.")
                elif host_stats["unknowwn"]:
                    ping_health.error(f"Ansible: {host} ping skipped.")
                else:
                    ping_health.warning(
                        f"Ansible: {host} status not understood: {ping_task_result_hosts[host]}."
                    )

        return ping_health


# pylint: disable=too-many-instance-attributes
class AnsiblePlaybookClientPlugin:
    """Ansible Playbook client plugin.

    Client plugin that allows control of and interaction with a ansible
    cluster using ansible-playbook cli

    ## Requirements

    1. this plugin uses subprocess to call a ansible binary, so you have to install
       ansible in the environment

    ## Usage

    """

    def __init__(
        self,
        environment,
        instance_id,
        inventory_path: str,
        ansiblecfg_path: str = "",
    ):
        """Initialize Ansible Playbook client.

        Parameters:
        -----------
        environment (Environment) : All metta plugins receive the environment
            object in which they were created.
        instance_id (str) : all metta plugins receive their own string identity.

        inventory_path (str) : string path to an inventory file.
        ansiblecfg_path (str) : string path to an optional ansiblecfg file.

        """
        self._environment: Environment = environment
        """Environemnt in which this plugin exists."""
        self._instance_id: str = instance_id
        """Unique id for this plugin instance."""

        self._ansibleplaybook = AnsiblePlaybookClient(
            ansiblecfg_path=ansiblecfg_path,
            inventory_path=inventory_path,
        )
        """Backend ansible-playbook cli handler."""

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Get info about a provisioner plugin."""
        return {
            "client": self._ansibleplaybook.info(deep=deep)
            if self._ansibleplaybook
            else "MISSING",
        }

    # pylint: disable=too-many-arguments
    def run_file(
        self,
        playbooksyml_path: str,
        extravars_path: str = "",
        extravars: Dict[str, str] = None,
        envs: Dict[str, str] = None,
        return_output: bool = False,
    ):
        """Run the ansible playbook install command on a yaml playbook file."""
        return self._ansibleplaybook.run(
            playbooksyml_path=playbooksyml_path,
            extravars_path=extravars_path,
            extravars=extravars,
            envs=envs,
            return_output=return_output,
        )



"""

Metta contrib package for using Mirantis launchpad to provision clusters.

This allows using launchpad to install the Mirantis products onto a cluster
that had already been provisioned by other services.

This module registers the package metta plugins.

"""
from typing import Any, Dict

from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.provisioner import METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER
from mirantis.testing.metta.client import METTA_PLUGIN_INTERFACE_ROLE_CLIENT
from mirantis.testing.metta_cli.base import METTA_PLUGIN_INTERFACE_ROLE_CLI

from .launchpad import LaunchpadClient
from .provisioner import (
    LaunchpadProvisionerPlugin,
    METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID,
    METTA_LAUNCHPAD_CONFIG_LABEL,
)
from .client import METTA_LAUNCHPAD_CLIENT_PLUGIN_ID, LaunchpadClientPlugin
from .cli import (
    LaunchpadCliPlugin,
    METTA_LAUNCHPAD_CLI_PLUGIN_ID,
)
from .launchpad import (
    METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT,
    METTA_LAUNCHPADCLIENT_WORKING_DIR_DEFAULT,
)


@Factory(
    plugin_id=METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID,
    interfaces=[
        METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER,
        METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID,
    ],
)
def metta_plugin_factory_provisioner_launchpad(
    environment: Environment,
    instance_id: str = "",
    label: str = METTA_LAUNCHPAD_CONFIG_LABEL,
    base: Any = LOADED_KEY_ROOT,
):
    """Create a launchpad provisioner plugin."""
    return LaunchpadProvisionerPlugin(environment, instance_id, label, base)


@Factory(
    plugin_id=METTA_LAUNCHPAD_CLIENT_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLIENT],
)
# pylint: disable=too-many-arguments
def metta_terraform_factory_client_launchpad(
    environment: Environment,
    instance_id: str = "",
    config_file: str = METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT,
    working_dir: str = METTA_LAUNCHPADCLIENT_WORKING_DIR_DEFAULT,
    cli_options: Dict[str, Any] = None,
    systems: Dict[str, Dict[str, str]] = None,
):
    """Create a launchpad client plugin."""
    return LaunchpadClientPlugin(
        environment=environment,
        instance_id=instance_id,
        config_file=config_file,
        working_dir=working_dir,
        cli_options=cli_options,
        systems=systems,
    )


@Factory(
    plugin_id=METTA_LAUNCHPAD_CLI_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI],
)
def metta_terraform_factory_cli_launchpad(environment: Environment, instance_id: str = ""):
    """Create a launchpad cli plugin."""
    return LaunchpadCliPlugin(environment, instance_id)


# ----- METTA BOOTSTRAPPERS -----


# pylint: disable=unused-argument
def bootstrap_environment(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added to.

    """



"""

Metta CLI : Fixture commands.

Cli plugin that allows examination of fixtures in environments.

"""

import logging
from typing import List

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures

from .base import CliBase, cli_output, METTA_PLUGIN_INTERFACE_ROLE_CLI

logger = logging.getLogger("metta.cli.fixtures")


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class FixturesCliPlugin(CliBase):
    """Fire command/group generator for fixture commands."""

    def fire(self):
        """Return a dict of commands."""
        return {"fixture": FixturesGroup(self._environment)}


class FixturesGroup:
    """Base Fire command group for fixtures commands."""

    def __init__(self, environment: Environment):
        """Attach environment to object."""
        self._environment: Environment = environment

    # pylint: disable=too-many-arguments
    def _filter(
        self,
        plugin_id: str = "",
        instance_id: str = "",
        interfaces: List[str] = None,
        labels: List[str] = None,
        skip_cli_plugins: bool = True,
    ):
        """Filter fixtures centrally."""
        matches = self._environment.fixtures().filter(
            plugin_id=plugin_id, instance_id=instance_id, interfaces=interfaces, labels=labels
        )

        if not skip_cli_plugins:
            return matches

        # filter out cli plugins
        fixtures = Fixtures()
        for fixture in matches:
            if METTA_PLUGIN_INTERFACE_ROLE_CLI not in fixture.interfaces:
                fixtures.add(fixture)
        return fixtures

    # methods needs to be on object for cli registation
    # pylint: disable=no-self-use
    def plugins(
        self,
        interface: str = "",
        label: str = "",
        skip_cli_plugins: bool = True,
    ):
        """List registered plugins and interfaces."""
        plugins_info = {}

        # use a protected property just for introspection
        # pylint: disable=protected-access
        for registration in Factory._registry.values():
            if skip_cli_plugins and METTA_PLUGIN_INTERFACE_ROLE_CLI in registration.interfaces:
                continue

            if interface and interface not in registration.interfaces:
                continue
            if label and label not in registration.labels:
                continue

            plugins_info[registration.plugin_id] = {
                "plugin_id": registration.plugin_id,
                "interfaces": registration.interfaces,
                "labels_from_factory": registration.labels,
            }

        return cli_output(plugins_info)

    # this is what is needed to limit or filter fixture introspection
    # pylint: disable=too-many-arguments
    def info(
        self,
        deep: bool = False,
        children: bool = False,
        plugin_id: str = "",
        instance_id: str = "",
        interface: str = "",
        label: str = "",
        skip_cli_plugins: bool = True,
    ):
        """Return Info for fixtures."""
        fixture_info_list = []
        for fixture in self._filter(
            plugin_id=plugin_id,
            instance_id=instance_id,
            interfaces=[interface] if interface else [],
            labels=[label] if label else [],
            skip_cli_plugins=skip_cli_plugins,
        ):
            fixture_info_list.append(fixture.info(deep=deep, children=children))

        return cli_output(fixture_info_list)

    # this is what is needed to limit or filter fixture introspection
    # pylint: disable=too-many-arguments
    def list(
        self,
        plugin_id: str = "",
        instance_id: str = "",
        interface: str = "",
        label: str = "",
        skip_cli_plugins: bool = True,
    ):
        """Return Info for fixtures."""
        fixture_info_list = []
        for fixture in self._filter(
            plugin_id=plugin_id,
            instance_id=instance_id,
            interfaces=[interface] if interface else [],
            labels=[label] if label else [],
            skip_cli_plugins=skip_cli_plugins,
        ):
            fixture_info_list.append(
                {
                    "plugin_id": fixture.plugin_id,
                    "instance_id": fixture.instance_id,
                    "interfaces": fixture.interfaces,
                    "labels": fixture.labels,
                }
            )

        return cli_output(fixture_info_list)



"""

Metta workload plugins that rely on ansiblecli clients.

Classes:
--------

AnsibleCliCoreWorkloadPlugin: Workload plugin for running ansible cli

AnsibleCliPlaybookWorkloadPlugin: Workload plugin for running ansible-playbook cli

"""
from typing import Dict, List, Any
import os

import yaml

from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures

from .ansiblecli_client import (
    AnsibleClientPlugin,
    AnsiblePlaybookClientPlugin,
    METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID,
    METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKCLIENT_PLUGIN_ID,
)

METTA_ANSIBLE_ANSIBLECLI_COREWORKLOAD_PLUGIN_ID = "metta_ansible_clicore_ansible_workload"
""" Ansible workload plugin id """

METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKWORKLOAD_PLUGIN_ID = (
    "metta_ansible_clicore_ansibleplaybook_workload"
)
""" Ansible-Playbook workload plugin id """

ANSIBLE_WORKLOAD_CONFIG_LABEL = "ansible"
""" default config label loading ansible config """

ANSIBLE_WORKLOAD_CONFIG_ARGS_KEY = "args"
"""Configures key for loading args."""
ANSIBLE_WORKLOAD_CONFIG_ENVS_KEY = "envs"
"""Configerus key for loading env variables to be used for clients."""
ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_KEY = "playbook.contents"
""" config key for the ansible playbook content """
ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_PATH_KEY = "playbook.path"
""" config key for the ansible playbook path to write to."""
ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_VARS_KEY = "vars.values"
""" config key for the ansible playbook vars."""
ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_VARS_PATH_KEY = "vars.path"
""" config key for the ansible playbook vars path to write to."""
ANSIBLE_WORKLOAD_CONFIG_ANSIBLECFG_KEY = "ansiblecfg.contents"
""" config key for the ansible cfg contents which will be written to file """
ANSIBLE_WORKLOAD_CONFIG_ANSIBLECFG_PATH_KEY = "ansiblecfg.path"
""" config key for the ansible cfg file path to write to."""
ANSIBLE_WORKLOAD_CONFIG_INVENTORY_KEY = "inventory.contents"
""" config key for the ansible  inventory, which will be passed to ansible """
ANSIBLE_WORKLOAD_CONFIG_INVENTORY_PATH_KEY = "inventory.path"
""" config key for the ansible inventory file path to write to."""


class AnsibleCliCoreWorkloadPlugin:
    """An ansiblecli workload that uses an AnsibleClientPlugin client."""

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        label: str = ANSIBLE_WORKLOAD_CONFIG_LABEL,
        base: Any = LOADED_KEY_ROOT,
    ):
        """Keep metta info and plugin config info."""
        self._environment: Environment = environment
        self._instance_id: str = instance_id

        self._config_label: str = label
        """ configerus load label that should contain all of the config """
        self._config_base: str = base
        """ configerus get key that should contain all plugin config """

        self._ansible_client: AnsibleClientPlugin = None
        """AnsibleClientPlugin which this plugin will use to interact with ansible."""

    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Get info about a provisioner plugin."""
        return {
            "config": {
                "label": self._config_label,
                "base": self._config_base,
            },
            "client-plugin": self._ansible_client.info(deep=deep)
            if self._ansible_client
            else "MISSING",
        }

    def prepare(self, fixtures: Fixtures = None):
        """Find the dependent fixtures."""
        if fixtures is None:
            fixtures = self._environment.fixtures()

        self._ansible_client = fixtures.get_plugin(
            plugin_id=METTA_ANSIBLE_ANSIBLECLI_CORECLIENT_PLUGIN_ID
        )

    def apply(self):
        """Apply the workload to the environment."""
        # Loaded configerus config for the plugin. Ready for .get().
        plugin_config = self._environment.config().load(self._config_label)

        args: List[str] = plugin_config.get([self._config_base, ANSIBLE_WORKLOAD_CONFIG_ARGS_KEY])
        envs: Dict[str, str] = plugin_config.get(
            [self._config_base, ANSIBLE_WORKLOAD_CONFIG_ENVS_KEY], default={}
        )

        return self._ansible_client.run(args=args, envs=envs)


class AnsibleCliPlaybookWorkloadPlugin:
    """An ansiblecli workload that uses an AnsiblePlaybookClientPlugin client."""

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        label: str = ANSIBLE_WORKLOAD_CONFIG_LABEL,
        base: Any = LOADED_KEY_ROOT,
    ):
        """Keep metta info and plugin config info."""
        self._environment: Environment = environment
        self._instance_id: str = instance_id

        self._config_label: str = label
        """ configerus load label that should contain all of the config """
        self._config_base: str = base
        """ configerus get key that should contain all plugin config """

        self._ansibleplaybook_client: AnsiblePlaybookClientPlugin = None
        """AnsiblePlaybookClientPlugin which this plugin will use to interact with ansible."""

    def info(self, deep: bool = False):
        """Get info about a provisioner plugin."""
        # Loaded configerus config for the plugin. Ready for .get().
        plugin_config = self._environment.config().load(self._config_label)

        return {
            "config": {
                "label": self._config_label,
                "base": self._config_base,
            },
            "plugin": {
                "playbook_contents": plugin_config.get(
                    [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_KEY], default={}
                ),
                "playbook_path": plugin_config.get(
                    [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_PATH_KEY],
                    default="",
                ),
                "envs": plugin_config.get(
                    [self._config_base, ANSIBLE_WORKLOAD_CONFIG_ENVS_KEY], default={}
                ),
                "vars_values": plugin_config.get(
                    [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_VARS_KEY], default={}
                ),
                "vars_path": plugin_config.get(
                    [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_VARS_PATH_KEY],
                    default="",
                ),
            },
            "client-plugin": self._ansibleplaybook_client.info(deep=deep)
            if self._ansibleplaybook_client
            else "MISSING",
        }

    def prepare(self, fixtures: Fixtures = None):
        """Find the dependent fixtures."""
        if fixtures is None:
            fixtures = self._environment.fixtures()

        self._ansibleplaybook_client = fixtures.get_plugin(
            plugin_id=METTA_ANSIBLE_ANSIBLECLI_PLAYBOOKCLIENT_PLUGIN_ID
        )

        # Loaded configerus config for the plugin. Ready for .get().
        plugin_config = self._environment.config().load(self._config_label)

        playbook_contents: str = plugin_config.get(
            [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_KEY], default={}
        )
        playbook_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_PATH_KEY],
            default="",
        )
        vars_values: str = plugin_config.get(
            [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_VARS_KEY], default={}
        )
        vars_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_VARS_PATH_KEY],
            default="",
        )

        if playbook_contents:
            os.makedirs(os.path.dirname(os.path.realpath(playbook_path)), exist_ok=True)
            with open(playbook_path, "w", encoding="utf8") as playbook_fileobject:
                yaml.safe_dump(playbook_contents, playbook_fileobject)
        else:
            if playbook_path and os.path.exists(playbook_path):
                os.remove(playbook_path)
        if vars_values:
            os.makedirs(os.path.dirname(os.path.realpath(vars_path)), exist_ok=True)
            with open(vars_path, "w", encoding="utf8") as vars_fileobject:
                yaml.safe_dump(vars_values, vars_fileobject)
        else:
            if vars_path and os.path.exists(vars_path):
                os.remove(vars_path)

    def apply(self):
        """Apply the workload to the environment."""
        # Loaded configerus config for the plugin. Ready for .get().
        plugin_config = self._environment.config().load(self._config_label)

        playbook_contents: str = plugin_config.get(
            [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_KEY], default={}
        )
        playbook_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_PATH_KEY],
            default="",
        )
        envs: Dict[str, str] = plugin_config.get(
            [self._config_base, ANSIBLE_WORKLOAD_CONFIG_ENVS_KEY], default={}
        )
        vars_values: str = plugin_config.get(
            [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_VARS_KEY], default={}
        )
        vars_path: str = plugin_config.get(
            [self._config_base, ANSIBLE_WORKLOAD_CONFIG_PLAYBOOK_VARS_PATH_KEY],
            default="",
        )

        if playbook_contents:
            os.makedirs(os.path.dirname(os.path.realpath(playbook_path)), exist_ok=True)
            with open(playbook_path, "w", encoding="utf8") as playbook_fileobject:
                yaml.safe_dump(playbook_contents, playbook_fileobject)
        else:
            if playbook_path and os.path.exists(playbook_path):
                os.remove(playbook_path)
        if vars_values:
            os.makedirs(os.path.dirname(os.path.realpath(vars_path)), exist_ok=True)
            with open(vars_path, "w", encoding="utf8") as vars_fileobject:
                yaml.safe_dump(vars_values, vars_fileobject)
        else:
            if vars_path and os.path.exists(vars_path):
                os.remove(vars_path)

        return self._ansibleplaybook_client.run_file(
            playbooksyml_path=playbook_path, extravars_path=vars_path, envs=envs
        )



"""

Metta Docker.

metta contrib package for docker functionality, specifically for registering
a Docker client plugin.

"""

from typing import Any

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.client import METTA_PLUGIN_INTERFACE_ROLE_CLIENT
from mirantis.testing.metta_health.healthcheck import METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK
from mirantis.testing.metta.workload import METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD

from mirantis.testing.metta_cli import METTA_PLUGIN_INTERFACE_ROLE_CLI

from .client import DockerPyClientPlugin, METTA_PLUGIN_ID_DOCKER_CLIENT
from .run_workload import (
    DockerPyRunWorkloadPlugin,
    METTA_PLUGIN_ID_DOCKER_RUN_WORKLOAD,
    DOCKER_RUN_WORKLOAD_CONFIG_LABEL,
    DOCKER_RUN_WORKLOAD_CONFIG_BASE,
)
from .cli import DockerCliPlugin, METTA_PLUGIN_ID_DOCKER_CLI


# This is really what it takes to configure both the metta plugin and the docker client
# pylint: disable=too-many-arguments
@Factory(
    plugin_id=METTA_PLUGIN_ID_DOCKER_CLIENT,
    interfaces=[
        METTA_PLUGIN_INTERFACE_ROLE_CLIENT,
        METTA_PLUGIN_ID_DOCKER_CLIENT,
        METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK,
    ],
)
def metta_plugin_factory_client_dockerpy(
    environment: Environment,
    instance_id: str = "",
    host: str = "",
    cert_path: str = "",
    tls_verify: bool = True,
    compose_tls_version: str = "TLSv1_2",
    version: str = "auto",
):
    """Create an metta client for dockerpy."""
    return DockerPyClientPlugin(
        environment,
        instance_id=instance_id,
        host=host,
        cert_path=cert_path,
        tls_verify=tls_verify,
        compose_tls_version=compose_tls_version,
        version=version,
    )


@Factory(
    plugin_id=METTA_PLUGIN_ID_DOCKER_RUN_WORKLOAD,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD],
)
def metta_plugin_factory_workload_docker_run(
    environment: Environment,
    instance_id: str = "",
    label: str = DOCKER_RUN_WORKLOAD_CONFIG_LABEL,
    base: Any = DOCKER_RUN_WORKLOAD_CONFIG_BASE,
):
    """Create a docker run workload plugin."""
    return DockerPyRunWorkloadPlugin(environment, instance_id, label=label, base=base)


@Factory(plugin_id=METTA_PLUGIN_ID_DOCKER_CLI, interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI])
def metta_docker_factory_cli(environment: Environment, instance_id: str = ""):
    """Create a docker cli plugin."""
    return DockerCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint METTA BootStrapping -----


# pylint: disable=unused-argument
def bootstrap_environment(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config
        added to.

    """



"""

Interact with the setuptools suite in a standardized way.

Functions:
----------

setuptools_entrypoint: run a setuptools entrypoint to bootstrap an argument
    using code provided by any python module.

"""
import logging
from typing import List, Dict, Any
from importlib import metadata

logger = logging.getLogger("metta.bootstrapp")

METTA_CONFIG_SETUPTOOLS_BOOTSTRAPS_KEY = "bootstraps"
"""Configerus .get() key for finding bootstrap entries."""


def setuptools_entrypoint(
    entrypoint: str, entries: List[str], args: List[Any], kwargs: Dict[str, Any]
):
    """Bootstrap some METTA objects with setuptools entrypoints.

    METTA bootstrapping is an attempt to allow an easy in to including
    contrib functionality without having to do a lot of Python imports.

    This is a setuptools enabled process, where any python package
    can declare a bootstraper, and this function will run that bootstrapper
    on request.
    The BootStrap entry_points are expected to receive a config object on
    which they can operate to add any specific or global functionality.

    BootStraps are typically used for two behaviours:

    1. just import files which run configerus or metta decorators to register
        plugins
    2. add source/formatter/validator plugins to the passed config object.

    Parameters:
    -----------
    bootstrap (List[str]) : a list of string bootstrapper entry_points for
        the ucct.bootstrap entry_points (part of setuptools.)
        Each value needs to refer to a valid entrypoint which will be
        executed with the config object as an argument.

    Raises:
    -------
    Raises a KeyError in cases of a bootstrap ID that cannot be found.

    Bootstrappers themselves may raise an exception.

    """
    for entry in entries:
        logger.debug("Running bootstrap entrypoint: %s=>%s ", entrypoint, entry)
        for metta_ep in metadata.entry_points()[entrypoint]:
            if metta_ep.name == entry:
                plugin = metta_ep.load()
                plugin(*args, **kwargs)
                break
        else:
            raise KeyError(f"Bootstrap not found {entrypoint}:{entry}")
